#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe de wrapping Scikit-Learn pour PyTorch

Created on Mon Oct 21 14:42:57 2019

@author: Cyrile Delestre
"""
import uuid
import pickle as pkl
from types import FunctionType
from typing import (Optional, ByteString, Union, Callable, Dict, Any, List,
                    Tuple)
from collections import OrderedDict
from warnings import warn

import numpy as np
from sklearn.base import BaseEstimator
import torch
from torch.optim import Optimizer
from torch.nn.modules.loss import _Loss
from torch.nn.utils import clip_grad_norm_
from torch.cuda.amp import GradScaler, autocast
from torch.utils.data import DataLoader, Dataset

from dstk.pytorch import (check_tensor, CallbackInterface, CallbackHandler,
                          FitState, FitControl, DEFAULT_CALLBACKS)
from dstk.utils.errors import DivergenceError

class BaseEnvironnement(BaseEstimator):
    r"""
    Classe d'environnement PyTorch pour être compatible Scikit-Learn.

    Notes
    -----
    Les attributs constituant BaseEnvironnement :
        _is_fitted : bool
            détermine si le modèle est entraîné

    Les méthodes constituant BaseEnvironnement :
        number_of_parameters : 
            nombre de paramètre (entraînable ou non) constiuant la 
            modélisation
        save_model : 
            sauvegarde du modèle via Pickle
        load_weights : 
            chargement des poids via Pickle. Si modèle non buildé, build le 
            modèle
        load_model :
            chargement d'un modèle à partir du Pickle au format save_weights
        fit :
            fonction d'apprentissage avec prototypage compatible avec 
            Scikit-Learn

    Les méthodes qui doivent être implémentées dans la classe du réseau :
        __init__ : 
            avec seulement les hyper-paramètres du réseau, afin de pouvoir 
            modifier les hyper-paramètres avec la méthode set_params de 
            BaseEstimator.
        build : 
            fonction d'implémentation des éléments du réseau PyTorch ainsi 
            que de l'attribut optimizer contenant l'optimizer du réseau
        forward :
            fonction PyTorch d'éxécution du réseau
        predict :
            fonction de prédiction, au prototypage Scikit-Learn 
            predict(X, **kargs) :
                - classifier : doit sortir une classe
                - regresseur : doit sortir une régression
        predict_proba : (optionnel)
            pour les classifier uniquement. Permet de sortir la probabilité 
            d'une observation dans chaqu'une des classes
        score : (optionnel) 
            au format de prototype Scikit-Learn. Si cette méthode n'est pas 
            implémentée par l'utilisateur, ce sont alors les méthodes par 
            défaut dans Scikit-Learn qui seront utilisées, à savoir dans le 
            cas d'un Regressor, R^2, et dans le cas de Classifier, l'accuracy.
    """
    _is_fitted: bool = False

    def build(self):
        r"""
        Méthode d'erreur si la fonction build n'a pas été implémentée dans 
        la classe modèle de l'utilisateur.
        """
        raise NotImplementedError(
            "La méthode build n'a pas été implémentée dans la classe "
            f"{self.__name__}. Se reporter à la documentation de "
            "BaseEnvironnement pour connaître les commodités d'implémentation "
            "de cette dernière."
        )

    def forward(self):
        r"""
        Méthode de forward du module de réseau profond PyTorch (voir la
        documention PyTorch).
        """
        raise NotImplementedError(
            "La méthode forward n'a pas été implémentée dans la classe "
            f"{self.__name__}. Se reporter à la documentation de "
            "BaseEnvironnement et PyTorch pour connaître les commodités "
            "d'implémentation de cette dernière."
        )

    def number_of_parameters(self, learn_params: bool=True):
        r"""
        Méthope retournant le nombre de paramètres présent dans le modèle.

        Parameters
        ----------
        learn_params : bool (=True)
            Compte les paramètres qui permettent l'apprentissage du modèle. 
            Sinon compte tous les paramètres (ceux qui sont dérivables et 
            non dérivables).
        """
        if learn_params:
            model_parameters = filter(
                lambda p: p.requires_grad, self.parameters()
            )
        else:
            model_parameters = self.parameters()
        return sum([np.prod(p.shape) for p in model_parameters])

    def save_model(
        self,
        path: Optional[str]=None,
        *,
        save_optimizer_state: bool=False,
        attr_list: Optional[Union[List[str], Tuple[str]]]=None,
        **kargs_dump
    ):
        r"""
        Méthode permettant de sauvegarder simplement les paramètres et les
        poids du modèle (sans la structure de réseau). Utilise Pickle.

        Parameters
        ----------
        path : Optional[str]
            path directory du modèle, si None retourne les données binaires de 
            Pickle.dumps()
        save_state_optimizer : bool
            booléen de sauvegarde de l'état de l'optimizer (par défaut False).
        attr_list : Optional[Union[List[str], Tuple(str)]]
            liste d'attributs à sauvegarder (par défaut None).
        **kargs_dump : 
            paramètres attachés à Pickle.dump(obj, file, **kargs_dump) si 
            path est un chemin ou a Pickles.dump(obj, **kargs_dump) si None.
        """
        # get_params est hérité de BaseEstimator
        params = self.get_params()
        state_dict = self.state_dict()
        if save_optimizer_state:
            state_optimizer = self.optimizer.state_dict()
        else:
            state_optimizer = None
        if attr_list:
            state_attribut = {
                attr: getattr(self, attr) for attr in attr_list
            }
        else:
            state_attribut = None
        if isinstance(path, str):
            pkl.dump(
                obj = [
                    params,
                    state_dict,
                    state_optimizer,
                    state_attribut
                ],
                file = open(path, 'bw'),
                **kargs_dump
            )
            return self
        elif path is None:
            return pkl.dumps(
                obj = [
                    params,
                    state_dict,
                    state_optimizer,
                    state_attribut
                ],
                **kargs_dump
            )
        else:
            raise AttributeError(
                "L'argument path doit être None ou un str et non de type "
                f"{type(path)}."
            )
        return self

    def load_weights(
        self, 
        path_or_bytes: Union[str, ByteString],
        **kargs_load
    ):
        r"""
        Méthode permettant de charger les poids et les paramètres du modèle
        via Pickle.

        Parameters
        ----------
        path_or_bytes : Union[str, ByteString]
            path directory du modèle ou bytes représentant le modèles
        **kargs_load :
            paramètres attachés à Pickle.load(file, **kargs_load) si path ou 
            Pickle.loads(data, **kargs_load) si bytes.
        """
        if not hasattr(self, 'build'):
            raise NotImplementedError(
                "Le modèle PyTorch n'a pas de méthode build (réf. doc de la "
                "classe)."
            )
        self.build()
        if isinstance(path_or_bytes, str):
            _, state_dict, state_optimizer, state_attribut = pkl.load(
                file = open(path_or_bytes, 'br'),
                **kargs_load
            )
        elif isinstance(path_or_bytes, bytes):
            _, state_dict, state_optimizer, state_attribut = pkl.loads(
                data = open(path_or_bytes, 'br'),
                **kargs_load
            )
        else:
            raise AttributeError(
                "path_or_bytes doit être soit de type str ou bytes et non "
                f"de type {type(path_or_bytes)}."
            )
        self.load_state_dict(state_dict);
        if state_optimizer:
            self.optimizer.load_state_dict(state_optimizer)
        if state_attribut:
            for key, item in state_attribut.items():
                setattr(self, key, item);
        self.eval()
        return self

    @classmethod
    def load_model(
        cls,
        path_or_bytes: Union[str, ByteString],
        **kargs_load
    ):
        r"""
        Fonction permettant de charger un modèle à partir d'une sauvegarde au
        format Pickle issue de save_weights(path).

        Parameters
        ----------
        path_or_bytes : Union[str, ByteString]
            path directory du modèle ou bytes représentant le modèles
        **kargs_load : 
            paramètres attachés à Pickle.load(file, **kargs_load) si path ou 
            Pickle.loads(data, **kargs_load) si bytes.

        Returns
        -------
        model : BaseEnvironnement
            modèle chargé
        """
        if isinstance(path_or_bytes, str):
            params, state_dict, state_optimizer, state_attribut = pkl.load(
                file = open(path_or_bytes, 'br'),
                **kargs_load
            )
        elif isinstance(path_or_bytes, bytes):
            params, state_dict, state_optimizer, state_attribut = pkl.loads(
                data = path_or_bytes,
                **kargs_load
            )
        else:
            raise AttributeError(
                "path_or_bytes doit être soit de type str ou bytes et non "
                f"de type {type(path_or_bytes)}."
            )
        model = cls(**params)
        if not hasattr(model, 'build'):
            raise NotImplementedError(
                "Le modèle PyTorch n'a pas de méthode build (réf. doc de la "
                "classe)."
            )
        model.build();
        model.load_state_dict(state_dict);
        if state_optimizer:
            model.optimizer.load_state_dict(state_optimizer)
        if state_attribut:
            for key, item in state_attribut.items():
                setattr(model, key, item)
        model.eval()
        return model

    def _extract_state(self, model):
        r"""
        Méthode privée qui extrait les paramètres du modèle et clone les 
        tensors PyTorch.

        Parameters
        ----------
        model :
            un modèle
        """
        return OrderedDict(
            (kk, vv.detach().clone()) for kk, vv in model.state_dict().items()
        )

    def _divergence(self, loss: float):
        r"""
        Méthode interne d'alerte de divergence du modèle pendant sa phase 
        d'apprentissage.
        """
        if np.isnan(loss) or np.isinf(loss):
            raise DivergenceError(
                """Verger ! L'algorithme a dit "verger" !"""
            )

    def _mod(self, n, mod):
        r"""
        Méthode interne permettant de calculer le modulo.
        """
        if mod:
            return n % mod
        else:
            return None

    def _train_step(
        self,
        data: List[Dict[str, Any]],
        loss_fn: Callable,
        loss_kargs: Dict[str, Any],
        state : FitState,
        control: FitControl,
        list_callbacks: List[CallbackInterface]
    ):
        r"""
        Méthode privée d'une étape de backpropagation du gradient.

        Parameters
        ----------
        data : List[Dict[str, Any]]
            Liste de dictionnaires sur les données d'un mini-batch.
        loss_fn : Callable
            Fonction coût PyTorch.
        loss_kargs : Dict[str, Any]
            Dictionnaire d'arguments à utiliser dans la fonction de coût 
            loss_fun.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        list_callbacks : List[CallbackInterface]
            Liste des callback permettant de contôler l'apprentissage durant 
            la phase d'évaluation.

        Returns
        -------
        loss : Valeur de la fonction coût au travers du mini-batch.
        """
        list_callbacks.begin_step(self, data, state, control, None)
        self.train()
        if control.device.type != "cpu":
            for key, value in data.items():
                if key not in control.expected_key_to_device:
                    data[key] = value.to(control.device)
        with autocast(enabled=control.mixed_type_gpu):
            output = self.forward(**data)
            target = check_tensor(data[control.target_field])
            loss_ = loss_fn(output, target, **loss_kargs)
        if control.gradient_accumulation > 1:
            loss_ /= control.gradient_accumulation
        control.grad_scaler.scale(loss_).backward()
        state.n_iter += 1
        if state.n_iter % control.gradient_accumulation == 0:
            if control.max_gradient_norm:
                control.grad_scaler.unscale_(self.optimizer)
                clip_grad_norm_(self.parameters(), control.max_gradient_norm)
            control.grad_scaler.step(self.optimizer)
            control.grad_scaler.update()
            self.optimizer.zero_grad()
        loss = loss_.detach()
        self._divergence(loss)
        list_callbacks.end_step(self, data, state, control, loss)
        return loss

    def _eval_step(
        self,
        data: List[Dict[str, Any]],
        loss_fn: Callable,
        loss_kargs: Dict[str, Any],
        control : FitControl
    ):
        r"""
        Méthode privée d'une étape deévaluation.

        Parameters
        ----------
        data : List[Dict[str, Any]]
            Liste de dictionnaires sur les données d'un mini-batch.
        loss_fn : Callable
            Fonction coût PyTorch.
        loss_kargs : Dict[str, Any]
            Dictionnaire d'arguments à utiliser dans la fonction de coût 
            loss_fun.
        control : FitControl
            Contrôle de l'apprentissage (instancé dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).

        Returns
        -------
        loss : Valeur de la fonction coût au travers du mini-batch.
        """
        self.eval()
        if control.device.type != "cpu":
            for key, value in data.items():
                if key not in control.expected_key_to_device:
                    data[key] = value.to(control.device)
        with torch.no_grad():
            output = self.forward(**data)
            target = check_tensor(data[control.target_field])
            loss = loss_fn(output, target, **loss_kargs)
        loss = loss.detach()
        self._divergence(loss)
        return loss

    def _iter_eval(
        self,
        eval_dataset: DataLoader,
        loss_fn: Callable,
        loss_kargs: Optional[Dict[str, Any]],
        state: FitState,
        control: FitControl,
        list_callbacks: List[CallbackInterface]
    ):
        r"""
        Méthode interne exécutant la phase d'évaluation durant l'apprentissage.
        Cette étape peut être appelé à la fin d'une époque ou durant 
        l'éxécution de l'époque (dépend de l'arguement iter_eval de la méthode 
        :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).

        Parameters
        ----------
        eval_dataset : DataLoader
            Dataset d'évaluation au format DataLoeader généré dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`.
        loss_fn : Callable
            Fonction coût à minimiser (utiliser les fonctions présentes dans 
            torch.nn.functional). Si la fonction est une fonction custom, elle 
            doit avoir le prototypage suivant :
                - loss_fn(y_estim, y_true, reduction='sum', **kargs)
            avec une implémentation de réduction 'sum' pour les phases 
            d'évaluation et 'mean' pour les phases d'apprentissage.
        loss_kargs : Optional[Dict[str, Any]]
            Dictionnaire d'arguments à utiliser dans la fonction de coût 
            loss_fun.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancé dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        list_callbacks : List[CallbackInterface]
            Liste des callback permettant de contôler l'apprentissage durant 
            la phase d'évaluation.
        """
        list_callbacks.begin_eval(
            model=self,
            data=eval_dataset,
            state=state,
            control=control,
            res_loss=None
        )
        loss_eval = 0
        for jj, eval_batch in enumerate(eval_dataset):
            list_callbacks.begin_eval_step(
                model=self,
                data=eval_batch,
                state=state,
                control=control,
                res_loss=loss_eval/(jj+1)
            )
            try:
                loss_eval += self._eval_step(
                    data=eval_batch,
                    loss_fn=loss_fn,
                    loss_kargs=loss_kargs,
                    control=control
                )
            except DivergenceError as error:
                control.diverge = True
                warn(error.msg)
                break
        list_callbacks.end_eval(
            model=self,
            data=eval_dataset,
            state=state,
            control=control,
            res_loss=loss_eval/(jj+1)
        )

    def fit(
        self,
        X: Union[List[Dict[str, Any]], Dataset, DataLoader],
        y: Optional[List[Union[int, float]]]=None,
        *,
        eval_dataset: Optional[DataLoader]=None,
        target_field: str='target',
        loss_fn: _Loss,
        loss_kargs: Dict[str, Any]=dict(),
        optimizer: Optional[Optimizer]=None,
        optimizer_kargs: Dict[str, any]=dict(),
        nb_epoch: int,
        iter_eval: Optional[int]=None,
        callbacks: List[CallbackInterface]=DEFAULT_CALLBACKS,
        device: str='cuda:0',
        mixed_type_gpu: bool=False,
        expected_key_to_device: List[str]=list(),
        gradient_accumulation: Optional[int]=None,
        max_gradient_norm: Optional[float]=None,
        dataloader_kargs: Dict[str, Any]=dict(),
        n_jobs: Optional[int]=None,
        state: Optional[FitState]=None,
        control: Optional[FitControl]=None,
        build: bool=True
    ):
        r"""
        Méthode de fitting généralisé du module PyTorch.
        
        Parameters
        ----------
        X : Union[List[Dict[str, Any]], Dataset, DataLoader]
            Dataset PyTorch contenant dans un dictionnaire les infos 
            nécessaires pour l'apprentissage du modèle.
        y : Optional[List[Union[int, float]]] 
            Pour être ISO avec Scikit-learn. L'ensemble de l'apprentissage
            passera via X et le système DataSet et DataLoader de PyTorch.
        eval_dataset : Optional[DataLoader]
            DataLoader PyTorch contenant les observations et les targets 
            d'évaluation. Si eval_dataset est None alors pas de early stopping 
            et évaluation des performances sur les données d'entrainement 
            (déconseillé car très fort risque d'overfitting).
        target_field : str (='target')
            Nom du champ contenu dans X de la target. Par défaut "target". 
            Attention l'implémentation de la foction collate_fn, peut faire 
            changer le nom de la target.
        loss_fn : _Loss
            Fonction coût à minimiser (utiliser les fonctions présentes dans 
            torch.nn.functional). Si la fonction est une fonction custom, elle 
            doit avoir le prototypage suivant :
                - loss_fn(y_estim, y_true, reduction='sum', **kargs)
            avec une implémentation de réduction 'sum' pour les phases 
            d'évaluation et 'mean' pour les phases d'apprentissage.
        loss_kargs : Dict[str, Any]
            Dictionnaire d'arguments à utiliser dans la fonction decoût 
            loss_fun. Si il n'y a pas d'arguments, mettre un dictionnaire 
            vide.
        optimizer : Optional[Optimizer]
            Objet optimizer de PyTorch. Si non renseigné doit être implémenté 
            dans le modèle et dans la partie build. Utile si un paramètre de 
            l'optimizer doit faire partie des hyperparamètres. Sinon à 
            renseigner ici pour alléger l'implémentation globale.
        optimizer_kargs: Dict[str, any]
            Paramètres pour l'optimizer, si optimizer inclue dans la 
            modélisation et l'argument optimizer est à None cet argument ne 
            sera pas pris en compte.
        nb_epoch : int (=5)
            Nombre d'époques max de l'apprentissage.
        iter_eval : Optional[int]
            Nombre d'itération entre deux évaluations. Actif seulement si 
            eval_set est non None. Si iter_eval est à None (définit par 
            défaut) les évaluations du modèle seront à la fin des époques. 
            Si un entier positif est renseigné les évaluations auront lieu 
            tous les iter_eval itérations d'apprentissage.
        callbacks : List[CallbackInterface] (=DEFAULT_CALLBACKS)
            Liste de callback intervenant aux différentes étapes de 
            l'entraînement et de l'évaluation. Pour avoir la liste des 
            différents callbacks implémentés par défaut voir 
            :func:`~dstk.pytorch._callback`. Par défaut utilise les 
            DEFAULT_CALLBACKS.
        device : str (='cuda:0')
            Indique le device sur le quel l'apprentissage s'exécute.
        mixed_type_gpu : bool (=False)
            Utilisation de nVidia AMP (Automatic Mixed Precision) pour GPU 
            permettant le typage mixte FP16 et FP32.
        expected_key_to_device : List[str] (=list())
            Liste des variables à exclure vers la charge sur GPU.
        gradient_accumulation : Optional[int]
            Nombre d'itération d'entrainement avec back propagation du 
            gradient avant d'appliquer la descente de gradient.
        max_gradient_norm : Optional[float]
            Valeur maximum du gradient, sinon clipping de ce dernier par la 
            valeur max_gradient_norm. Si non renseigné alors pas de gradient 
            clipping. Le clipping du gradient peut être utilie pour l'approche 
            par accumulation des gradient.
        dataloader_kargs : Dict[str, Any] (=dict())
            Si X n'est pas un objet Dataloader permet d'entrer les arguments 
            du Dataloader (voir la `documention PyTorch_ `pour connaitre les 
            arguments du Dataloader).
        n_jobs : Optional[int]
            Nombre de process des étapes d'inférence et de backpropagation des 
            gradients :
                - Si y est not None :
                    forcé à 1 ;
                - Si None :
                    utilise automatiquement le nombre de processeurs 
                    (et pas le nombre de threads) ;
                - Si un entier :
                    utilise le nombre de processus indiqué.
        state : Optional[FitState]
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : Optional[FitControl]
            Contrôle de l'apprentissage (instancé dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        build : bool (=True)
            Permet de reconstruire le réseau si, entre la définition de la 
            classe et le fit, il y a changement d'hyper-paramètres (exemple : 
            RandomizedSearchCV ou toute autre fonction Scikit-learn pouvant 
            utiliser set_params).

        Warnings
        --------
        Si le but du modèle est d'être utilisé avec d'autre module 
        Scikit-Learn il ne faut pas que le type de X soit un DataLoader 
        (les autres types reste valide). De manière général, pour la méthode 
        fit, il est conseillé d'utiliser le type DataSet PyTorch comme type 
        d'entré à X.

        .. _documention PyTorch: https://pytorch.org/docs/stable/data.html
        """
        if not hasattr(self, 'optimizer') and optimizer is None:
            raise NotImplementedError(
                "L'estimateur ne possède par d'objet 'optimizer'. Cette "
                "attribue doit être un objet d'otimisation de type PyTorch : "
                "torch.optim.(SGD, ADAM, etc.). Il faut le déclarer dans "
                "la construction de la classe modèle ou le définir dans "
                "les arguments de la méthode fit."
            )

        if iter_eval is not None and iter_eval < 0:
            raise AttributeError(
                "iter_eval doit être soit None ou un entier positif."
            )

        if (
                gradient_accumulation is not None and
                iter_eval < gradient_accumulation
        ):
            warn(
                "iter_eval < gradient_accumulation donc à chaque itération "
                "un pas de gradient ne sera pas effectué."
            )

        if optimizer is not None:
            if hasattr(self, 'optimizer'):
                warn(
                    "Un optimizer a été donné en argument de la méthode fit "
                    "hors il en existait déjà un attribue optimizer dans la "
                    "classe du modèle. La priorité est accordé à celui en "
                    "entré du fit et à donc écrasé celui du modèle de base."
                )
            self.optimizer = optimizer(self.parameters(), **optimizer_kargs)

        list_callbacks = CallbackHandler(callbacks)
        list_callbacks.copy()

        id_processes = str(uuid.uuid1())

        # Permet de construire le modèle avec les paramètres actuels du fit.
        if build:
            self.build()

        if not isinstance(X, DataLoader):
            if isinstance(X, dict):
                X = [X]
            dataset = DataLoader(
                dataset=X,
                **dataloader_kargs
            )
        else:
            dataset = X

        # Si y est non None c'est que le processes passe par un search de
        # Scikit-Learn.
        if y is not None:
            torch.set_num_threads(1)
            dataset.num_workers = 0
        elif n_jobs is not None:
            torch.set_num_threads(n_jobs)
        else:
            torch.set_num_threads(torch.get_num_threads())

        if not state:
            state = FitState(
                id_process=id_processes,
                nb_epoch=nb_epoch,
                epoch_max_steps=len(dataset),
                eval_max_steps=len(eval_dataset) if eval_dataset else 0,
                loss_name=(
                    loss_fn.__name__ if isinstance(loss_fn, FunctionType)
                    else loss_fn.__class__.__name__
                ),
                n_iter=0
            )
        if not control:
            control = FitControl(
                target_field=target_field,
                device=(
                    device := torch.device(
                        device if torch.cuda.is_available() else 'cpu'
                    )
                ),
                mixed_type_gpu=mixed_type_gpu and device.type != "cpu",
                grad_scaler=GradScaler(
                    enabled=(mixed_type_gpu and device.type != "cpu")
                ),
                expected_key_to_device=expected_key_to_device,
                gradient_accumulation=gradient_accumulation,
                max_gradient_norm=max_gradient_norm
            )
        self.to(control.device)

        list_callbacks.begin_train(self, dataset, state, control, None)
        try:
            for n_epoch in range(nb_epoch):
                state.n_epoch = n_epoch
                loss_epoch = 0
                list_callbacks.begin_epoch(
                    model=self,
                    data=dataset,
                    state=state,
                    control=control,
                    res_loss=loss_epoch
                )
                for ii, batch in enumerate(dataset):
                    if (
                        eval_dataset and
                        self._mod(state.n_iter, iter_eval) == 0
                    ):
                        self._iter_eval(
                            eval_dataset=eval_dataset,
                            loss_fn=loss_fn,
                            loss_kargs=loss_kargs,
                            state=state,
                            control=control,
                            list_callbacks=list_callbacks
                        )
                        if control.diverge or control.training_stop:
                            break
                    try:
                        loss_step = self._train_step(
                            data=batch,
                            loss_fn=loss_fn,
                            loss_kargs=loss_kargs,
                            state=state,
                            control=control,
                            list_callbacks=list_callbacks
                        )
                        loss_epoch += loss_step
                    except DivergenceError as error:
                        control.diverge = True
                        warn(error.msg)
                        break
                loss_epoch /= (ii+1)
                list_callbacks.end_epoch(
                    model=self,
                    data=dataset,
                    state=state,
                    control=control,
                    res_loss=loss_epoch
                )
                if eval_dataset and iter_eval is None:
                    self._iter_eval(
                        eval_dataset=eval_dataset,
                        loss_fn=loss_fn,
                        loss_kargs=loss_kargs,
                        state=state,
                        control=control,
                        list_callbacks=list_callbacks
                    )
                if control.diverge or control.training_stop:
                    break
        except (Exception, KeyboardInterrupt) as e:
            list_callbacks.error_interrupt(
                model=self,
                data=dataset,
                state=state,
                control=control,
                res_loss=None,
                exception=e
            )
            raise
        list_callbacks.end_train(self, dataset, state, control, None)
        if control.device.type != "cpu": self.to(torch.device('cpu:0'))
        self._is_fitted = True
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe de wrapping Scikit-Learn pour PyTorch. Base regressor.

Created on Mon Oct 21 14:42:57 2019

@author: Cyrile Delestre
"""
from typing import Union, Dict, List, Tuple, Any, Optional, Iterable

import numpy as np
from pandas.core.frame import DataFrame
from sklearn.base import RegressorMixin
from torch import no_grad
from torch.utils.data import Dataset

from ._base import BaseEnvironnement
from . import check_tensor_dict, check_tensor

class BaseRegressor(BaseEnvironnement, RegressorMixin):
    r"""
    Classe d'implémentation des Regressor pour être compatible Scikit-Learn. 
    Elle hérite de BaseEnvironnement et RegressorMixin.

    Notes
    -----
    La méthode constituant BaseRegressor :
        predict:
            méthode de prédiction compatible avec le prototypage Scikit-Learn 
            predict(X, rnn_hidden_state, **kargs).
    """
    _estimator_type = 'regressor'
    def predict(
        self,
        X: Union[Dataset,
                 np.ndarray,
                 DataFrame,
                 List[Dict[str, Any]],
                 Tuple[Dict[str, Any]],
                 Dict[str, Any]],
        nan_inf: bool=True,
        fill_nan: float=0,
        add_dim_batch: bool=False,
        **kargs
    ):
        r"""
        Méthode de prédiction compatible Scikit-Learn.
        
        Parameters
        ----------
        X : Union[Dataset, np.ndarray, DataFrame, List[Dict[str, Any]], Tuple[Dict[str, Any]], Dict[str, Any]]
            Features du modèle, pour le bon fonctionnement via Scikit-Learn et
            de son utilisation classique, X peut être de différents types
                - Dataset: Classe Dataset de PyTorch, doit sortir un 
                    dictionnaire contenant tous les arguments indispensables au
                    calcul du modèle de la fonction forward.
                - list: liste de dictionnaire contenant tous les arguments
                    indispensables au calcul du modèle de la fonction forward.
            Attention : La première dimension doit être consacrée à celle du
            batch. Si qu'une seule observation alors la première dimension sera
            de dimension 1.
        nan_inf : bool
            traitre les NaN si l'algorithme à divergé et met les valeurs inf 
            à la plus grande valeur possible du type du numpy array.
        fill_nan : float
            valeur de remplacement des NaN.
        add_dim_batch : bool (=False)
            ajoute la dimension du batch si elle n'est pas présente.

        Returns
        -------
        res : np.ndarray
            retourne le résultat au format numpy array de la fonction de 
            classification forward
        """
        self.eval()
        with no_grad():
            if isinstance(X, (Dataset, list, tuple,)):
                res = []
                for data in X:
                    data = check_tensor_dict(data, add_dim_batch=add_dim_batch)
                    res.append(self.forward(**data).detach().numpy())
                res = np.concatenate(res, axis=0)
                return np.nan_to_num(res, nan=fill_nan) if nan_inf else res
            elif isinstance(X, np.ndarray):
                res = self.forward(check_tensor(X), **kargs).detach().numpy()
                return np.nan_to_num(res, nan=fill_nan) if nan_inf else res
            elif isinstance(X, dict):
                X = check_tensor_dict(X, add_dim_batch=add_dim_batch)
                res = self.forward(**X).detach().numpy()
                return np.nan_to_num(res, nan=fill_nan) if nan_inf else res
            else:
                raise AttributeError(
                    f"Le type de X {type(X)} n'est pas prise en compte."
                )

class BaseRegressorOnline(BaseRegressor):
    r"""
    Classe d'implémentation des Regressor dans un contexte online pour être 
    compatible à Scikit-Learn. Elle hérite de BaseRegressor.
    
    Notes
    -----
    Les méthodes constituant BaseRegressorOnline :
        save_weights :
            surchage de la méthode save_weights de BaseEnvironnement afin de 
            coller au contexte online.
        fit :
            surcharge de la méthode fit de BaseEnvironnement afin de coller au 
            context online.
        fit_online :
            implémentation de l'apprentissage online avec possibilité de 
            réaliser une prédiction à la volée.
    """
    def save_model(self, path: Optional[str]=None, **kargs_dump):
        r"""
        Méthode permettant de sauvegarder simplement les paramètres et les
        poids du modèle (sans la structure de réseau) ainsi que l'état de 
        l'optimizer pour une utilisation online. Utilise Pickle.
        
        Parameters
        ----------
        path : Optional[str]
            path directory du modèle, si None retourne les données binaires de 
            Pickle.dumps()
        **kargs_dump : 
            paramètres attachés à Pickle.dump(obj, file, **kargs_dump) si 
            path est un chemin ou a Pickles.dump(obj, **kargs_dump) si None.
        """
        return super().save_model(
            path=path,
            save_optimizer_state=True,
            **kargs_dump
        )

    def fit(self, X, y=None, **kargs):
        r"""
        Méthode de fit (wrapper vers Scikit-learn). Pour l'apprentissage la
        partie preprocessing doit être faite au préalable. La fonction est
        conçue pour fonctionner avec du multiprocessing. Donc un identifiant
        unique est donné à chaque session d'entrainement via uuid avec pour
        protocole 1.
        
        Parameters
        ----------
        X : Iterable
            Dataset PyTorch contenant dans un dictionnaire les infos 
            nécessaires pour l'apprentissage du modèle.
                - Attention :
                    à la sortie du DataLoader (donc via collate_fn ou 
                    directement via DataSet de PyTorch), il faut que le 
                    dictionnaire possède les noms présents dans le forward de 
                    la sous-classe du modèle.
        y : Optional[Iterable]
            Pour être ISO avec Scikit-learn. L'ensemble de l'apprentissage 
            passera via X et le système DataSet et DataLoader de PyTorch.
        **kargs :
            le reste des arguments de la méthode fit de la classe 
            BaseEnvironnement.
        """
        super().fit(
            X=X,
            y=y,
            eval_dataset=None,
            nb_epoch=1,
            kargs_dataloader=dict(
                batch_size=1,
                shuffle=False,
                drop_last=False
            ),
            callbacks=[],
            **kargs
        )
        return self

    def fit_online(
        self,
        X: Iterable,
        y: Optional[Iterable]=None,
        predict: bool=False,
        needs_proba: bool=False,
        gap: int=0,
        treat_nan_inf: bool=True,
        fill_nan: float=0,
        **fit_kargs
    ):
        r"""
        Méthode d'apprentissage online avec possibilité d'inférence à la volée.
        
        Parameters
        ----------
        X : Iterable
            Dataset PyTorch contenant dans un dictionnaire les infos
            nécessaires pour l'apprentissage du modèle.
                - Attention :
                    à la sortie du DataLoader (donc via collate_fn 
                    ou directement via DataSet de PyTorch), il faut que le 
                    dictionnaire possède les noms présents dans le forward de 
                    la sous-classe du modèle.
        y : Optional[Iterable]
            Pour être ISO avec Scikit-learn. L'ensemble de l'apprentissage 
            passera via X et le système DataSet et DataLoader de PyTorch.
        predict : bool
            permet de faire la phase d'inférence à la volée. Par défaut False.
        gap : int
            Permet de définir un gap entre l'apprentissage du modèle et la 
            prédiction. Par défaut 0, aucun gap. Dans la pratique le gap 
            est à minima égal à l'horizon de prédiction
        treat_nan_inf : bool
            traitre les NaN de la prédiction si l'algorithme à divergé et met 
            les valeurs inf à la plus grande valeur possible du type du numpy 
            array.
        fill_nan : float
            valeur de remplacement des NaN pour la prédiction.
        **fit_kargs :
            le reste des arguments de la méthode fit de la classe 
            BaseEnvironnement.
        """
        if 'build' in fit_kargs.keys():
            fit_kargs_ = dict(fit_kargs)
            if fit_kargs_['build']:
                fit_kargs_['build'] = False
                self.build()
        else:
            fit_kargs_ = dict(fit_kargs)
            fit_kargs_['build'] = False
        if predict:
            res = []
        for ii, data in enumerate(X):
            self.fit(data, y=y, **fit_kargs_)
            if predict:
                if gap > 0:
                    if ii+gap < len(X):
                        data_eval = X[ii+gap]
                        res.append(
                            self.predict(
                                data_eval,
                                treat_nan_inf=treat_nan_inf,
                                fill_nan=fill_nan,
                                add_dim_batch=True
                            )
                        )
                else:
                    res.append(
                        self.predict(
                            data,
                            treat_nan_inf=treat_nan_inf,
                            fill_nan=fill_nan,
                            add_dim_batch=True
                        )
                    )
        if predict:
            return np.concatenate(res, axis=0)
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe pour client SSH.

Created on Sat Apr 11 06:41:28 2020

@author: Cyrile Delestre
"""

import logging
from os.path import dirname
from typing import List, Dict, Optional
from functools import reduce
from subprocess import run as sp_run

class ClientSSH:
    r"""
    Classe de connextion SSH permettant d'envoyer, exécuter et récupérer des 
    données sur un serveur distant (typiquement les bastions). La classe a 
    pour optique de faciliter le lancement de scipts Pig directement d'une 
    machine locale.
    
    Parameters
    ----------
    hostname : Optional[str]
        nom du hostname du serveur distant. Si cette valeur n'est pas 
        renseigné alors un promp pour demander le serveur souvrira dans 
        le terminal.
    username : Optional[str]
        nom de l'utilisateur pour identification. Si cette valeur n'est pas 
        renseigné alors uin prompt pour demander l'identifiant de 
        l'utilisateur souvrira dans le terminal.
    ssh_key : Optional[str]
        chemin vers la clef SSH privé. Si non reseigné un prompt s'ouvrira 
        pour demande le mot de passe associé au username.
    timeout : int
        timeout des commande SSH. ATTENTION : Une commande Pig peut être long. 
        Si pendant l'exécution le timeout est atteint et qu'il y a d'autre 
        instruction dans le script il y aura une erreur Python. Mais la 
        procédure Pig continura en fond sur la cellule Hadoop, sauf si le 
        terminal avec la session Python ce ferme.
    verbose : bool
        affiche le retour de la commande SSH.
    
    Notes
    -----
    La procédure standars pour lancer un scipt Pig est :
        
        1 : Envoyer le Pig sur un des bastions
        
        2 : Envoyer les données nécéssaire pour l'exécution du Pig sur 
        Bastion
        
        3 : Envoyer les données sur HDFS
        
        4 : Exécuter le scipt Pig
        
        5 : Récupérer les résultat sous Bastion
        
        6 : Les renvoyer en local
        
        7 : Supprimer les données de Bastion
        
    ou une alternative à l'étape 6 et 7 :
        
        5 bis : Envoyer sur staging
        
        6 bis : Récupérer les données directement depuis HDFS via le 
        staging
    
    Il existe dans la classe des procédures aggregant certaines de ces étapes.
    
    Si aucun ssh_key n'est renseigné alors un mot de passe sera demandé. Dans 
    un terminal bash ce mot de passe sera masqué, mais ce ne sera pas le cas 
    dans un terminal iPython (incompatible avec le masquage des inputs).
    
    Examples
    --------
    L'exemple suiviant consiste à mettre l'ensemble des fichiers jar et 
    script sur un serveur distant afin de permettre l'exécusion d'un script 
    pig. La commande pig serait équivalente à :
    ::
        pig -useHCatalog -f script.pig -p param_1="param_1" -p param_2=42 -p param_3="param_3"
    >>> from dstk.ssh import ClientSSH
    >>> 
    >>> # conf_ssh dictionnaire contenant les paramètres de connexions SSH
    >>> ssh = ClientSSH(**conf_ssh, verbose=False, timeout=10800)
    >>> 
    >>> # SCP d'un fichier vers un serveur distant
    >>> ssh.cp(local_path_file, server_path_file)
    >>> 
    >>> # SCP de plusieur script Pig vers un serveur distant
    >>> for script in all_script_path:
    >>>     ssh.cp(f"{local_path_script}{script}", server_path_file)
    >>> 
    >>> # SCP d'un repertoir local ver un serveur distant
    >>> ssh.cp(local_path_folder, server_path_folder, recursive=True)
    >>> 
    >>> # Upload sur la cellule Hadoop d'un fichier nécessaire à l'exécution 
    >>> # d'un script Pig
    >>> ssh.hadoop_put(f"{server_path}/data_input.csv")
    >>> 
    >>> # Exécution d'un script Pig qui utilise HCatalog pour une table Hive
    >>> ssh.pig(
    >>>     f"{server_path}/script.pig",
    >>>     params=dict(
    >>>         param_1='param_1',
    >>>         param_2=42,
    >>>         param_3='param_3'
    >>>     ),
    >>>     opt=['useHCatalog']
    >>> )
    >>> 
    >>> # Déconnexion au serveur
    >>> ssh.logout()
    
    See also
    --------
    download_file
    """
    def __init__(self,
                 hostname: Optional[str]=None,
                 username: Optional[str]=None,
                 ssh_key: Optional[str]=None,
                 timeout: int=30,
                 verbose: bool=False):
        try:
            global getpass
            import getpass
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Le module getpass n'est pas installé : l'instruction "
                "import getpass renvoit une erreur "
                "ModuleNotFoundError. Rajouter getpass dans les "
                "packages à installer dans le fichier environment.yml du "
                "projet."
            )
        try:
            global pxssh
            from pexpect import pxssh
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Le module pexpect n'est pas installé : l'instruction "
                "from pexpect import pxssh renvoit une erreur "
                "ModuleNotFoundError. Rajouter pexpect dans les "
                "packages à installer dans le fichier environment.yml du "
                "projet."
            )
        self.hostname = input('hostname: ') if hostname is None else hostname
        self.username = input('username: ') if username is None else username
        self.ssh_key = ssh_key
        self.timeout = timeout
        self.verbose = verbose
        self._connexion()

    def _connexion(self):
        try:
            self.conn = pxssh.pxssh(timeout=self.timeout)
            if self.ssh_key is None:
                password = getpass.getpass('password: ')
                self.conn.login(
                    server = self.hostname,
                    username = self.username,
                    password = password
                )
            else:
                self.conn.login(
                    server = self.hostname,
                    username = self.username,
                    ssh_key = self.ssh_key
                )
        except pxssh.ExceptionPxssh as e:
            print("La connexion pxssh a échouer.")
            print(e)

    def test_connexion(self):
        r"""
        Méthode permettant de tester la connexion SSH. Retourne la commande 
        'uptime'. Si la connexion est correctement établie retourne sur le 
        terminal le uptime du serveur, sinon retourne une erreur.
        """
        self.exe_cmd('uptime', True)
        return self

    def exe_cmd(self, cmd: str, verbose: Optional[bool]=None):
        r"""
        Méthode d'éxécution d'une commande sur le serveur distant.
        
        Parameters
        ----------
        cmd : str
            cammande (shell, bach, etc.) à exécuter.
        verbose : Optional[bool]
            affiche le retour de la commande. Si None utilise le paramétrage 
            par défaut de la classe ClientSSH.
        
        Examples
        --------
        La commande suivant retourne les fichiers contenue dans le repertoire 
        courant :
        >>> from dstk.ssh import ClientSSH
        >>> 
        >>> # conf_ssh dictionnaire contenant les paramètres de connexions SSH
        >>> ssh = ClientSSH(**conf_ssh, verbose=False, timeout=10800)
        >>> ssh.exe_cmd('ls', verbose=True)
        """
        verbose = verbose if verbose is not None else self.verbose
        self.conn.sendline(cmd)
        self.conn.prompt()
        if verbose:
            print(self.before())
        return self

    def before(self):
        r"""
        Permet de rendre interprétable dans Python les retours des prompts 
        ce produisant avec le timeout.
        """
        return self.conn.before.decode('utf-8').replace('\r\n', '\n')

    def after(self):
        r"""
        Permet de rendre interprétable dans Python les retours des prompts 
        ce produisant après le timeout.
        """
        return self.conn.after.decode('utf-8').replace('\r\n', '\n')

    def check_cmd(self):
        r"""
        Permet de récupérer le sortie d'une commande.
        """
        self.conn.sendline("echo $?")
        self.conn.prompt()
        return self.before().split('\n')[1]

    def pig(self,
            path_script: str,
            params: Dict[str, str]=dict(),
            opt: List[str]=[]):
        r"""
        Méthode d'exécution d'un script Pig.
        
        Parameters
        ----------
        path_scipt : str
            chemin sur le serveur distant du scipt Pig à exécuter
        params : Dict[str, str]
            dictionnaire de l'ensemble des paramètres à envoyer au srcipt Pig.
        opt : List[str]
            list d'option Pig.
        
        Examples
        --------
        L'exemple suivant est equivalent à lancer manuellement sur le serveur 
        distant la commande :
        ::
            pig -useHCatalog -f script.pig -p param_1="param_1" -p param_2=42 -p param_3="param_3"
        
        >>> from dstk.ssh import ClientSSH
        >>> 
        >>> # conf_ssh dictionnaire contenant les paramètres de connexions SSH
        >>> ssh = ClientSSH(**conf_ssh, verbose=False, timeout=10800)
        >>> ssh.pig(
        >>>     "script.pig",
        >>>     params=dict(
        >>>         param_1='param_1',
        >>>         param_2=42,
        >>>         param_3='param_3',
        >>>     ),
        >>>     opt=['useHCatalog']
        >>> )
        """
        query = "pig"
        if len(opt) > 0:
            for op in opt:
                query += f" -{op}"
        query += f" -f {path_script}"
        if len(params) > 0:
            for kk, vv in params.items():
                query += f' -p {kk}=\"{vv}\"'
        self.exe_cmd(query)
        return self

    def hadoop_put(self,
                   path_file_server: str,
                   path_file_hadoop: Optional[str]=None,
                   verbose: Optional[bool]=None,
                   if_exist: bool=True):
        r"""
        Permet de mettre un fichier du serveur distant vers la cellule 
        Hadoop.
        
        Parameters
        ----------
        path_file_server : str
            chemin du fichier sur le serveur distant.
        path_file_hadoop : Optional[str]
            chemin de destination sur la cellule Hadoop.
        verbose : Optional[bool]
            affiche le retour de commandes.
        if_exist : bool
            créé le répertoire sur la cellule Hadoop sur ce dernier n'éxiste 
            pas.
        """
        if path_file_hadoop is None:
            path_file_hadoop = "./"
        if if_exist:
            # Check si le réperoire existe
            path_hadoop = '/'.join(path_file_hadoop.split('/')[:-1])+'/'
            exist = (
                self.exe_cmd(f"hadoop fs -ls {path_hadoop}", False)
                .check_cmd()
            )
            if exist != '0':
                logging.warning(
                    f"Le répertoire {path_hadoop} sur la cellule Hadoop "
                    "n'existe pas. Création de cette dernière."
                )
                self.exe_cmd(f"hadoop fs -mkdir {path_hadoop}", False)
            # Check si le fichier existe
            file_name = path_file_hadoop.split('/')[-1]
            if file_name == '':
                file_name = path_file_server.split('/')[-1]
            exist = (
                self.exe_cmd(f"hadoop fs -ls {path_hadoop}{file_name}", False)
                .check_cmd()
            )
            if exist == '0':
                logging.warning(
                    f"Le fichier {file_name} existe déjà. "
                    "Suppression de ce dernier."
                )
                self.exe_cmd(f"hadoop fs -rmr {path_hadoop}{file_name}", False)
        query = f"hadoop fs -put {path_file_server} {path_file_hadoop}"
        self.exe_cmd(query, verbose)
        return self

    def hadoop_getmerge(self,
                        path_file_hadoop: str,
                        path_file_server: Optional[str]=None,
                        check_is_directory: bool=False,
                        verbose: Optional[bool]=None):
        r"""
        Permet de faire un getmerge de la cellule Hadoop vers un répertoire 
        du serveur distant.
        
        Parameters
        ----------
        path_file_hadoop : str
            chemin vers le répertoire Hadoop où sont stockés tous les segments 
            d'un même fichier. Fonctionne également si le chemin pointe vers 
            un simple fichier.
        path_file_server : Optional[str]
            chemin vers le fichier du serveur distant vers le quel seront 
            fusionné tous les sergments.
        check_is_directory : bool
            check si le chemin entré est un répertoire ou un fichier (par 
            défaut False). Pour une exécution plus rapide sans passer par ce 
            test finir le path_file_or_folder_hadoop par "/*" si on pointe 
            vers un répertoire évitant ainsi ce test.
        verbose : Optional[bool]
            affiche le retour de commandes.
        
        Notes
        -----
        Il est déconseillé généralement de passer par cette méthode de 
        récupération de donnée si le fichier en trop volumineux pouvant 
        saturer très rapidement l'espace disque du serveur distant. Pour les 
        gros fichier il est conseiller de stocker sur le répertoire staging 
        de la cellule et de passer par get_url_staging qui permet de créer 
        le chemin URL du fichier et d'utiliser 
        :func:`~dstk.ssh._utils.download_file`. Celà permet de passer de 
        Hadoop vers un machine local sans passer par le serveur distant.
        
        See also
        --------
        get_url_staging, :class:`~dstk.ssh._utils.download_file`
        """
        siz_file = 0
        if check_is_directory:
            query = f"hadoop fs -ls {path_file_hadoop}/"
            self.exe_cmd(query, False)
            siz_file = int(
                list(
                    filter(
                        lambda x: len(x) > 0,
                        self.before().split('\n')[1].split(' ')
                    )
                )[4]
            )
        query = f"hadoop fs -getmerge {path_file_hadoop}"
        if siz_file > 0 and check_is_directory:
            query += "/*"
        query = (
            query if path_file_server is None
            else query + f" {path_file_server}"
        )
        self.exe_cmd(query, verbose)
        return self

    def get_url_staging(self,
                        path_staging: str,
                        file_name: Optional[str]=None,
                        content_length: bool=True):
        r"""
        Méthode permettant de récuper l'URL d'un fichier (ou repertoire si 
        le fichier est segmenté) sur le staging de la cellule Hadoop.
        
        Parameters
        ----------
        path_staging : str
            chemin vers le répertoire du staging.
        file_name : Optional[str]
            nom du fichier que l'on souhaite récupérer (attention si None 
            descend l'ensemble de répertoire).
        content_length : bool
            Permet d'intégrer la taille du fichier dans l'URL permettant 
            d'avoir une bare de progression du téléchargement de la donnée 
            via :class:`~dstk.ssh._utils.download_file`.
        
        Notes
        -----
        Cette méthode est à utiliser avec le fonction 
        :func:`~dstk.ssh._utils.download_file` qui, a partir de l'URL généré, 
        télécharger la donnée depuis la cellule Hadoop directement vers 
        l'ordinateur local.
        
        See also
        --------
        :class:`~dstk.ssh._utils.download_file`
        """
        if not path_staging.startswith("/hdfs/staging/out/"):
            path = f"/hdfs/staging/out/{path_staging}"
        else:
            path = path_staging
        info = self.exe_cmd(f"hadoop fs -ls {path}").before()
        if self.check_cmd() != '0':
            raise IOError(f"Fichier {path} inexistant.")
        url = (
            "http://hadoop-manny-staging.s.arkea.com:8080/hadoop-staging/get"
            + path
        )
        if file_name is not None or content_length:
            url += "?"
        if file_name is not None:
            url += f"file-name={file_name}"
        if content_length:
            info = list(
                map(
                    lambda x: list(
                        filter(
                            lambda y: len(y) > 0,
                            x.split(' ')
                        )
                    ),
                    filter(
                        lambda x: len(x) > 0,
                        info.split("\n")
                    )
                )
            )
            length = reduce(lambda x,y: int(y[4])+x, info[1:], 0)
            if file_name is not None:
                url += "&"
            url += f"content-length={length}"
        return url

    def cp(self,
           file_path_local: str,
           file_path_server: Optional[str]=None,
           recursive: bool=False,
           if_exist: bool=True,
           verbose: Optional[bool]=None):
        r"""
        Méthode SCP (Secure CoPy) d'une machine local vers un serveur distant.
        
        Parameters
        ----------
        file_path_local : str
            chemin vers un fichier ou un répertoire local.
        file_path_server : Optional[str]
            chemin vers un répertoire sur le serveur distant.
        recursive : bool
            si file_path_local pointe vers un répertoire, permet de copié 
            récursivement les fichiers et répertoires contenu dans le 
            répertoire vers le serveur distant.
        if_exist : bool
            permet de tester si le répertoire sur le serveur distant existe, 
            s'il n'éxiste pas le crée.
        verbose : Optional[bool]
            affiche le retour de commandes.
        """
        verbose = verbose if verbose is not None else self.verbose
        option = "rp" if recursive else "p"
        if file_path_server is None:
            path_server = "./"
        else:
            path_server = file_path_server
        if if_exist:
            dir_server = dirname(file_path_server)
            exist = self.exe_cmd(f"ls {dir_server}").check_cmd()
            if exist == '2':
                logging.warning(
                    f"Le répertoire {dir_server} n'existe pas. "
                    "Création de ce dernier."
                )
                self.exe_cmd(f"mkdir {dir_server}")
        if isinstance(file_path_local, list):
            dquote = '"'
            file_path_local = f"{dquote}{' '.join(['a', 'b'])}{dquote}"
            path_server = dirname(file_path_server)
        query = [
            "scp",
            f"-{option}",
            file_path_local,
            f"{self.username}@{self.hostname}:{path_server}"
        ]
        output = sp_run(query)
        if verbose:
            print(output.stdout)
        return self

    def logout(self):
        r"""
        Déconnexion du client SSH.
        """
        self.conn.logout()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Outils associés à la gestion des métriques

Created on Sun Nov 22 12:32:15 2020

@author: Cyrile Delestre
"""

import logging
from typing import Any, Dict

def _push_logger(logger: logging.Logger, message: Dict[str, Any]):
    r"""
    Fonction privée de push vers un logger. Si None print le résultat.
    """
    if logger is None:
        print(message, flush=True)
    elif isinstance(logger, logging.Logger):
        logger.info(message)
    elif isinstance(logger, list):
        logger.append(message)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classes pour simplifier l'exploitation des bases SQL par Pandas.

Created on Mon Nov 23 09:23:52 2020

@author: Cyrile Delestre
"""

import sqlite3
from dataclasses import dataclass, field
from typing import Dict, Any, Union
from warnings import warn

import pandas as pd

@dataclass
class PandasSQL:
    r"""
    Classe de manipulation Pandas pour attaquer une base SQL. Gère toutes 
    les connexions aux bases de données compatibles avec la méthode Pandas
    pandas.read_sql_query :
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_sql_query.html

    Parameters
    ----------
    conn : sqlite3.Connection
        connecteur vers une base SQL.

    Notes
    -----
    Attention si vous utilisez la méthode close() le connecteur sera perdu. 
    Il fraudra alors réinstancier la classe PandasSQL avec un nouveau 
    connecteur.

    Attention cette classe n'est pas Picklelisable, car les connecteurs sont 
    des objets non binarisables. Dans ce cas utiliser PandasSQLite ou 
    PandasSQLAlchemy avec un context manager avec le mot clef "with".

    Examples
    --------
    >>> import sqlite3
    >>> from dstk.data import PandasSQL

    Si on considère une base SLQLite3 "base.db" dans le répertoire data du 
    projet et que cette base possède une table "base_name".

    >>> df = PandasSQL(sqlite3.connect("data/base.db"))
    >>> df("SELECT * FROM base_name;")
       field_name_1  field_name_2
    0             d             1
    1             e             2
    2             f             3

    >>> gen = df.gen_data("SELECT * FROM base_name;")
    >>> type(gen)
    >>> generator
    >>> next(gen)
    ('d', 1)

    >>> df.close()
    >>> df("SELECT * FROM base_name;")
    AttributeError: L'instance PandasSQL n'a pas de connecteur. 
    Ceci est peut-être du au fait que vous avez fermé la connexion avec la 
    méthode close.

    La classe PandasSQL n'est pas binarisable, elle est donc inutilisable 
    dans un context multi-processing (voir PandasSQLite ou PandasSQLAlchemy).

    >>> import pickle
    >>> df = PandasSQL(sqlite3.connect("data/base.db"))
    >>> pickle.dumps(df)
    TypeError: can't pickle sqlite3.Connection objects

    See also
    --------
    PandasSQLite, PandasSQLAlchemy
    """
    conn: sqlite3.Connection

    def __call__(self, cmd: str, **sql_ope):
        r"""
        Le call est un alise de read_data
        """
        return self.read_data(cmd=cmd, **sql_ope)

    def _conn_test(self):
        r"""
        Méthode privée de test de connexion.
        """
        if not hasattr(self, 'conn'):
            raise AttributeError(
                "L'instance PandasSQL n'a pas de connecteur. Ceci est "
                "peut-être du au fait que vous avez fermé la connexion "
                "avec la méthode close."
            )

    def read_data(self, cmd: str, **sql_opt):
        r"""
        Permet de lire les données à partir de la commande SQL cmd.

        Parameters
        ----------
        cmd : str
            commande SQL
        **sql_opt :
            arguments de pandas.read_sql_query

        Returns
        -------
        :df: DataFrame Pandas
        """
        self._conn_test()
        return pd.read_sql_query(cmd, self.conn, **sql_opt)
    
    def gen_data(self, cmd: str):
        r"""
        Générateur de la commande SQL. Attention en fonction du connecteur 
        cette méthode peut ne pas fonctionner. Il faudra alors la surcharger.

        Parameters
        ----------
        cmd : str
            commande SQL

        Returns
        -------
        :data: générateur
        """
        self._conn_test()
        cur = self.conn.cursor()
        cur.execute(cmd)
        gen = (data for data in cur.fetchall())
        cur.close()
        return gen

    def close(self):
        r"""
        Permet de fermer le connecteur SQL. Attention en fonction du 
        connecteur cette méthode peut ne pas fonctionner. Il faudra alors la 
        surcharger.
        """
        self._conn_test()
        self.conn.commit()
        self.conn.close()
        # del important sinon objet non Picklelisable
        del self.conn
    

class PandasSQLite(PandasSQL):
    r"""
    Classe de manipulation Pandas pour attaquer une base SQL. Utilise le 
    moteur SQLite3 de Python. Elle n'est compatible que SQLite mais ne 
    nécessite pas d'installation de package supplémentaire.

    Parameters
    ----------
    path_sql : Union[str, sqlite3.Connection]
        chemin vers une base SQLite3 (utilisation recommendé).
        Peut être également un connecteur SQLite3.
    kargs_sql : 
        arguments d'entrée au connecteur SQLite3 :
            sqlite3.connect(path_sql, **kargs_sql)

    Notes
    -----
    Contrairement à PandasSQL cette classe est binarisable car ne fait pas 
    d'effet de bord sur les connecteurs. Il faut l'utiliser dans un contexte 
    manager avec le mot clef "with".

    Attention si vous utilisez un connecteur SQLite3 a la place de d'un 
    chemin vers la base, l'objet n'est plus binarisable !

    Examples
    --------
    >>> from dstk.data import PandasSQLite

    Si on considère une base SLQLite3 "base.db" dans le répertoire data du 
    projet.

    >>> df = PandasSQLite("data/base.db")
    >>> 
    >>> with df:
    >>>     info = df.info_database()
    >>> info
    {'base_name': {'field_name_1': 'INTEGER'}, {'field_name_2': 'TEXT'}}

    >>> with df:
    >>>     out = df("SELECT * FROM base_name;")
    >>> out
       field_name_1  field_name_2
    0            31       janvier
    1            29       février
    2            31          mars

    >>> with df:
    >>>     gen = df.gen_data("SELECT * FROM base_name;")
    >>> type(gen)
    generator
    >>> next(gen)
    (31, 'janvier')

    Cette classe est bien binarisable.

    >>> import pickle
    >>> pickle.dumps(df)
    b'...'

    See also
    --------
    PandasSQL, PandasSQLAlchemy
    """ 
    def __init__(self, path_sql: Union[str, sqlite3.Connection], **kargs_sql):
        if isinstance(path_sql, sqlite3.Connection):
            self.conn = path_sql
            self.path_sql = path_sql
            warn(
                "Attention ! Vous utilisez comme arguement path_sql un "
                "connecteur SQLite3. Ce n'est pas l'utilisation recommendé. "
                "Vous ne pourrez par exemple pas paralléliser un process "
                "avec cette instanciation de PandasSQLite !"
            )
        else:
            self.path_sql = path_sql
        self.kargs_sql = kargs_sql
    
    def __enter__(self):
        r"""\
        Contexte de PandasSQL.
        """
        if not isinstance(self.path_sql, sqlite3.Connection):
            self.conn = sqlite3.connect(self.path_sql, **self.kargs_sql)
        return self

    def __exit__(self, error_type, error_value, error_traceback):
        r"""\
        Sortie du contexte.
        """
        if not isinstance(self.path_sql, sqlite3.Connection):
            self.close()

    def get_cursor(self):
        r"""\
        Renvoit un cursor si la connexion avec la base est établie.
        """
        self._conn_test()
        return self.conn.cursor()

    def info_database(self):
        r"""\
        Retourne sous un dictionnaire les informations de la base.
        
        Returns
        -------
        out : dict
            dict au format {table_name: {field_name: field_type}}
        """
        self._conn_test()
        cur = self.conn.cursor()
        info = {
            tt[0]: {
                ii[1]: ii[2] 
                for ii in cur.execute(
                    f"PRAGMA TABLE_INFO({tt[0]});"
                ).fetchall()
            }
            for tt in cur.execute(
                "SELECT name FROM sqlite_master WHERE type=='table';"
            ).fetchall()
        }
        cur.close()
        return info

class PandasSQLAlchemy(PandasSQL):
    r"""
    Classe de manipulation Pandas pour attaquer une base SQL. Utilise le 
    moteur SQLAlchemy. L'avantage de cette librairie est d'être compatible 
    avec beaucoup de bases de données.

    Parameters
    ----------
    path_sql : str
        URL de connexion SQLAlchemy. L'URL de connexion est définit par :
            "dialect+driver://usernam:password@host:port/database"
    kargs_sql : Dict[str, Any]
        arguments d'entrée au connecteur SQLAlchemy :
            sqlalchemy.create_engine(path_sql, **kargs_sql)

    Notes
    -----
    Contrairement à PandasSQL cette classe est binarisable car ne fait pas 
    d'effet de bord sur les connecteurs. Il faut l'utiliser dans un contexte 
    manager avec le mot clef "with".

    Cette classe nécessite SQLAlchemy, si ce package n'est pas présent dans 
    votre environnement il suffit d'ajouter le package "sqlalchemy" au fichier 
    environment.yml et de lancer le script bash bash_install_env.sh dans le 
    répertoire shell du projet.

    Examples
    --------
    >>> from dstk.data import PandasSQLAlchemy

    Si on considère une base SLQLite3 "base.db" dans le répertoire data du 
    projet.

    >>> df = PandasSQLAlchemy("sqlite:///data/base.db")
    >>> 
    >>> with df:
    >>>     info = df.info_database()
    >>> info
    {'base_name': {'field_name_1': 'TEXT'}, {'field_name_2': 'TEXT'}}

    >>> with df:
    >>>     out = df("SELECT * FROM base_name;")
    >>> out
       field_name_1  field_name_2
    0          riri        picsou
    1          fifi        donald
    2        loulou        mickey

    >>> with df:
    >>>     gen = df.gen_data("SELECT * FROM base_name;")
    >>> type(gen)
    generator
    >>> next(gen)
    ('riri', 'picsou')

    Cette classe est bien binarisable.

    >>> import pickle
    >>> pickle.dumps(df)
    b'...'

    See also
    --------
    PandasSQL, PandasSQLite
    """ 
    path_sql: str
    kargs_sql: Dict[str, Any] = field(default_factory=dict)
    
    def __init__(self, path_sql: str, **kargs_sql):
        try:
            global sqlalchemy
            import sqlalchemy
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Le module SQLAlchemy n'est pas installé : l'instruction "
                "import sqlalchemy renvoit une erreur "
                "ModuleNotFoundError. Rajouter sqlalchemy dans les "
                "packages à installer dans le fichier environment.yml du "
                "projet."
            )
        self.path_sql = path_sql
        self.kargs_sql = kargs_sql

    def __enter__(self):
        r"""
        Contexte de PandasSQL.
        """
        self.engine =  sqlalchemy.create_engine(
            self.path_sql,
            **self.kargs_sql
        )
        self.conn = self.engine.connect()
        return self

    def __exit__(self, error_type, error_value, error_traceback):
        r"""
        Sortie du contexte.
        """
        self.close()

    def get_engine(self):
        r"""
        Méthode qui permet de récupérer l'engine SQLAlchemy. Attention il 
        faut que la connection se fasse explicitement dans la classe 
        (path_sql non None et engine = 'sqlalchemy'.
        """
        if hasattr(self, 'engine'):
            return self.engine
        raise AttributeError(
            "Il n'y a pas d'engine SQLAlchemy dans cette instance de "
            "PandasSQLAlchemy."
        )

    def info_database(self):
        r"""
        Retourne sous un dictionnaire les informations de la base.
        
        Returns
        -------
        out : dict
            dict au format {table_name: {field_name: field_type}}
        """
        self._conn_test()
        import sqlalchemy
        metadata = sqlalchemy.MetaData()
        metadata.reflect(bind = self.engine)
        info = {
            tt: {
                ii.name: str(ii.type) 
                for ii in ff.columns
            }
            for tt, ff in metadata.tables.items()
        }
        return info

    def gen_data(self, cmd: str):
        r"""
        Générateur de la commande SQL.

        Parameters
        ----------
        cmd : str
            commande SQL

        Returns
        -------
        data : generator
            générateur sur la data en sortie de la commande appliqué sur la 
            base.
        """
        self._conn_test()
        res = self.conn.execute(cmd)
        return (data for data in res.fetchall())

    def close(self):
        r"""\
        Permet de fermer le connecteur SQL.
        """
        self._conn_test()
        self.conn.close()
        self.engine.dispose()
        # del important sinon objet non Picklelisable
        del self.engine, self.conn



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe et fonction facilitant le wrapping Scikit-Learn d'un transformer 
personalisé.

Created on Mon Nov 23 12:12:30 2020

@author: Cyrile Delestre
"""

from collections.abc import Iterable, Callable

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from joblib import Parallel, delayed
from tqdm import tqdm

def make_transformer(udf_transform: Callable,
                     n_jobs: int=1,
                     verbose: bool=False):
    r"""
    Instentiatieur de classe Transformer, sert à éviter d'écrire une classe 
    directement.
    
    Parameters
    ----------
    udf_transform : Callable
        fonction de transformation défini par l'utilisateur (user difined 
        function) qui doit avoir pour prototypage :
            udf_transform(X, **kargs)
    n_jobs : int
        nombre de processeurs en parallèle
    verbose : int
        Si True bare de progression de process.
    
    Notes
    -----
    Si le transformer que l'on souhaite créer nécessite la surcharge de la 
    fonction fit alors il est obliger de passer par la définition d'une 
    classe.
    
    De même udf_transform doit être une fonction simple, sinon il est très 
    conseiller de passer par la définition d'une classe.
    
    Examples
    --------
    >>> from dataclasses import dataclass
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.pipeline import Pipeline
    >>> from dstk.features import make_transformer
    >>> from dstk.utils import check_dataframe
    >>> 
    >>> def udf_transform(X, order_col=['col_3', 'col_1', 'col_2', 'col_0']):
    >>>     # check_dataframe vérifie s'il s'agit bien d'une DataFrame 
    >>>     # et que les colonnes de order_col sont bien présentes
    >>>     check_dataframe(X, order_col)
    >>>     return X[order_col]
    >>> 
    >>> orderer = make_transformer(udf_transform)
    >>> clf = RandomForestClassifier()
    >>> pipe = Pipeline([('order_col', orderer), ('rf', clf)])
    >>> pipe.fit(X, y)
    
    See also
    --------
    Transformer
    """
    if not isinstance(udf_transform, Callable):
        raise AttributeError(
            "udf_transform doit être une fonction Callable et non de type "
            f"{type(udf_transform)}."
        )
    trans = Transformer(n_jobs=n_jobs, verbose=verbose)
    trans.udf_transform = udf_transform
    return trans

class Transformer(BaseEstimator, TransformerMixin):
    r"""
    Classe générique de transformation Scikit-Learn qui implémente une 
    exécution compatible avec les Pipeline où les éléments doivent être 
    traités un à un si ils sont dans une liste ou un itérable. 
    
    Parameters
    ----------
    n_jobs : int
        nombre de processeurs en parallèle
    verbose : bool
        Si True bare de progression de process.
    
    Notes
    -----
    Il est important d'implémenter la méthode udf_transform(self, X, **kargs)  
    et surcharger les autres  méthodes si besoin.
    
    Examples
    --------
    Exemple très simple d'un transformer utilisant la classe Transfomer qui 
    permet d'ordonner les colonnes d'une DataFramme en fonction d'un ordre 
    souhaité.
    
    >>> from dataclasses import dataclass
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.pipeline import Pipeline
    >>> from dstk.features import Transformer
    >>> from dstk.utils import check_dataframe
    >>> 
    >>> @dataclass
    >>> class SortColumns(Transformer):
    >>>     "Classe qui ordonne les colonnes d'une DataFrame."
    >>>     order_col: list
    >>>     n_jobs: int=1
    >>>     verbose: bool=False
    >>>
    >>>     def udf_transform(self, X):
    >>>         # check_dataframe vérifie s'il s'agit bien d'une DataFrame 
    >>>         # et que les colonnes de self.order_col sont bien présentes
    >>>         check_dataframe(X, self.order_col)
    >>>         return X[self.order_col]
    
    >>> orderer = SortColumns(['col_3', 'col_1', 'col_2', 'col_0'], n_jobs=2)
    >>> clf = RandomForestClassifier()
    >>> pipe = Pipeline([('order_col', orderer), ('rf', clf)])
    >>> pipe.fit(X, y)
    
    See also
    --------
    make_transformer
    """
    def __init__(self, n_jobs: int, verbose: bool):
        self.n_jobs = n_jobs
        self.verbose = n_jobs

    def fit(self, X: Iterable, y: None=None):
        r"""
        Fonction fit pour être ISO avec Scikit-Learn
        
        Parameters
        ----------
        X : Iterable
            Observations
        y : None
            pour être ISO avec Scikit-Learn
        """
        return self

    def _transform_iter(self, X: Iterable, **kargs):
        r"""
        Itération de transformations.
        
        Parameters
        ----------
        X : Iterable
            Observations
        """
        iterator = tqdm(
            X, 
            desc = f"{self.__class__.__name__}",
            mininterval = 0.5
        ) if self.verbose else X

        trans = delayed(self.udf_transform)
        with Parallel(n_jobs=self.n_jobs) as par:
            return par(trans(x, **kargs) for x in iterator)

    def udf_transform(self, X: Iterable, **kargs):
        r"""\
        Transformation unitaire.
        
        Parameters
        ----------
        X : Iterable
            Observations
        """
        raise NotImplementedError(
            "Il faut implémenter la méthode udf_transform(self, X, **kargs) "
            f"à la classe {self.__name__} qui hérite de la classe "
            "Transformer."
        )

    def transform(self, X: Iterable, **kargs):
        r"""\
        Méthode générique scikit-learn pour la transformation.

        Parameters
        ----------
        X : Iterable
            Array-like, pour un traitement itératif un Iterable quelconque
        :**kargs: arguments propres au transformer
        """
        if isinstance(X, (pd.DataFrame, np.ndarray,)):
            return self.udf_transform(X, **kargs)
        elif isinstance(X, Iterable):
            return self._transform_iter(X, **kargs)
        else:
            raise AttributeError(
                "L'argument X de transform doit être un Iterable."
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Toutes les classes erreurs personalisées
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Created on Thu Nov 26 12:15:33 2020

@author: Cyrile Delestre
"""

from typing import Optional

class NoConsistentError(Exception):
    r"""
    Classe d'erreur de consistance (représentation matricielle non dense).
    
    Parameters
    ----------
    msg : Optional[str]
        message à passer avec l'erreur.
    """
    def __init__(self, msg: Optional[str]=None):
        self.msg = msg

class CheckError(Exception):
    r"""
    Classe d'erreur des ckeckage.
    
    Parameters
    ----------
    msg : Optional[str]
        message à passer avec l'erreur.
    """
    def __init__(self, msg: Optional[str]=None):
        self.msg = msg

class DivergenceError(Exception):
    r"""
    Classe d'erreur de convergence.
    
    Parameters
    ----------
    msg : Optional[str]
        message à passer avec l'erreur.
    """
    def __init__(self, msg: Optional[str]=None):
        self.msg = msg

class TrainError(Exception):
    r"""
    Classe d'erreur quelconque durant l'apprentissage.
    
    Parameters
    ----------
    msg : Optional[str]
        message à passer avec l'erreur.
    """
    def __init__(self, msg: Optional[str]=None):
        self.msg = msg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe aggrégatrice des mesures.

Created on Sun Nov 22 12:41:19 2020

@author: Cyrile Delestre
"""

from sys import version_info

import numpy as np
import pandas as pd

if version_info.major >= 3 and version_info.minor >= 8:
    class Aggregator:
        r"""
        Classe agrégatrice pour le décorateur 
        :func:`~dstk.metrics._metrics.sniffer` offrant un comportement par 
        défaut. Il est possible de faire hériter cette classe pour construire 
        d'autres agrégateurs ou changer son comportement.
        
        Elle doit contenir la méthode make_message qui renvoit un 
        dictionnaire. Rien n'y oblige, mais en standard elle renvoit tourjours 
        au moins le type de la variable avec le mot clef 'type'.
        
        Examples
        --------
        Voici un exemple de surchage de la classe Aggregator où l'on souhaite 
        changer le comportement sur les listes et ajouter un nouveau type 
        MonObj.
        
        >>> from functools import wraps, singledispatchmethod
        >>> from dstk.metrics import Aggregator
        >>> 
        >>> class MonAggregator(Aggregator):
        >>>     "docstring de la classe"
        >>> 
        >>>     # Important : redéfinir la méthode make_message en la décorant
        >>>     # de singledispatchmethod et qui renvoit vers le make_message
        >>>     # de la classe mère
        >>>     @singledispatchmethod
        >>>     def make_message(self, input):
        >>>         return super().make_message(input)
        >>> 
        >>>     # Ajout d'un agrégateur pour le type MonObj
        >>>     @make_message.register
        >>>     def _(self, input: MonObj):
        >>>         return dict(type='MonObj', out=input.out)
        >>> 
        >>>     # Surchage du type list
        >>>     @make_message.register
        >>>     def _(self, input: list):
        >>>         return dict(
        >>>             type = 'list',
        >>>             len = len(input),
        >>>             out = input[0]
        >>>          )
        """
        from functools import singledispatchmethod
        
        @singledispatchmethod
        def make_message(self, input):
            return dict(
                type='unknown'
            )

        @make_message.register
        def _(self, input: str):
            return dict(
                type='str',
                out=input
            )

        @make_message.register(float)
        @make_message.register(np.floating)
        @make_message.register(np.float)
        @make_message.register(int)
        @make_message.register(np.integer)
        def _(self, input):
            return dict(
                type='scalar',
                out=input
            )
    
        @make_message.register(list)
        @make_message.register(tuple)
        def _(self, input):
            if not np.isscalar(input[0]):
                return dict(
                    type='list',
                    inside=str(type(input[0])),
                    len=len(input)
                )
            else:
                if len(input) > 10:
                    return dict(
                        type='list',
                        inside='scalar',
                        len=len(input)
                    )
                else:
                    info = dict(
                        type='list',
                        len=len(input),
                        inside='scalar'
                    )
                    info.update(
                        {
                            f"out_{ii}": ss for ii, ss in enumerate(input)
                        }
                    )
                    return info
    
        @make_message.register
        def _(self, input: np.ndarray):
            return dict(
                type='array',
                mean=np.mean(input),
                std=np.std(input),
                dim=len(input.shape)
            )
    
        @make_message.register
        def _(self, input: pd.DataFrame):
            info = dict(
                type='dataframe',
                nb_row=input.shape[0],
                nb_col=input.shape[1]
            )
            info.update(
                {
                    f"{col}_mean": mean
                    for mean, col in zip(input.mean(), input.columns)
                }
            )
            info.update(
                {
                    f"{col}_std": std
                    for std, col in zip(input.std(), input.columns)
                }
            )
            return info
    
        @make_message.register
        def _(self, input: dict):
             return dict(type='dict', dict=input)
    
        @make_message.register
        def _(self, input: None):
            return dict(type="none")
else:
    from warnings import warn
    warn(
        "La classe Aggregator pour le décorateur sniffer a besoin de "
        "Python 3.8 minimum et actuellement vous tournez sur la version "
        f"{version_info.major}.{version_info.minor}. Le sniffer n'est pas "
        "inutilisable, mais il vous faudra concevoir un Aggregator."
    )
    class Aggregator:
        r"""
        Classe coquille si la version de Python est inférieure à 3.8.
        """
        def make_message(self, input):
            warn(
                "Classe Aggregator coquille (vide), la classe n'est pas "
                "implémentée."
            )
            return dict(
                type = 'pas de type',
                msg = 'Aggregator à implémenter'
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jan  9 11:59:50 2021

@author: cyriledelestre
"""

from abc import ABCMeta

class SuperclassDocstring(ABCMeta):
    r"""
    Méta classe permettant de faire hériter la documention d'une calsse 
    abstraite à la classe concrète.
    """
    def __new__(mcls, classname, bases, cls_dict):
        cls = ABCMeta.__new__(mcls, classname, bases, cls_dict)
        mro = cls.__mro__[1:]
        for name, member in cls_dict.items():
            if not getattr(member, '__doc__'):
                for base in mro:
                    try:
                        member.__doc__ = getattr(base, name).__doc__
                        break
                    except AttributeError:
                        pass
        return cls


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Fonctions ou classes utiles pour le la visualisation et l'analyse.

Created on Tue Apr 14 22:44:11 2020

@author: Cyrile Delestre
"""

from typing import Callable, Optional, NewType
from math import erf

import numpy as np

TypeCDF = NewType(
    'TypeCDF',
    Callable[[np.ndarray, Optional[float], Optional[float]], np.ndarray]
)

@np.vectorize
def cdf_normal(x: np.ndarray, mean: float=0, std: float=1):
    r"""
    Fonction vectorisée permettant de calculer la CDF de la loi Normale.
    
    Parameters
    ----------
    x : ndarray
        points d'évaluation de la CDF.
    mean : float
        moyenne
    std : float
        écart-type
    
    See also
    --------
    cdf_rayleigh, kolmogorov_smirnov_test
    """
    return 1/2*(1+erf((x-mean)/(std*np.sqrt(2))))

def cdf_rayleigh(x: np.ndarray, mean: float=1, std: None=None):
    r"""
    Fonction vectorisée permettant de calculer la CDF de la loi Rayleigh.
    
    Parameters
    ----------
    x : ndarray
        points d'évaluation de la CDF.
    mean : float
        moyenne
    std : None
        pour être ISO avec la fonction 
        :func:`~dstk.visualization.statistique.cdf_normal` (inutilisé).
    
    See also
    --------
    cdf_normal, kolmogorov_smirnov_test
    """
    if mean <= 0:
        raise AttributeError(
            "La moyenne de la loi de Rayleigh est strictement positive."
        )
    x_ = np.zeros_like(x)
    filt = x >= 0
    sig = mean*(np.pi/2)**2
    x_[filt] = 1 - np.exp(-x[filt]**2/(2*sig**2))
    return x_

def kolmogorov_smirnov_test(x: np.ndarray,
                            cdf: TypeCDF,
                            mean: Optional[float],
                            std: Optional[float],
                            **kargs):
    r"""
    Fonction permettant de calculer l'écart entre deux CDF du test de 
    Kolmogorow-Smirnov.
    
    Parameters
    ----------
    x : np.ndarray
        variable aléatoire suivant une loi inconnue.
    cdf : TypeCDF
        fonction de CDF continue.
    mean : Optional[float]
        moyenne estimée de la variable x.
    std : Optional[float]
        écrat-type estimé de la variable x.
    **kargs :
        arguments pour la fonction histogram de Numpy.
    
    Returns
    -------
    Ecart de Kolmogorow-Smirnov.
    
    Notes
    -----
    La fonction cdf doit avoir le format :
        cdf(x: np.array, mean: float, std: float)
    Les arguments mean et std sont juste des noms en référence à la loie 
    normale, elles peuvent représentrer d'autres carastéristiques 
    statistiques. Si la cdf à besoins de plus de 2 caractéristiques pour être 
    calculé, cette fonction ne fonctionnera pas.
    
    .. todo:: Faire une méthode qui permettrait d'avoir autant de 
        caractéristiques statistiques souhaité. (attention a RocAnalyser !)
    
    See also
    --------
    cdf_normal, cdf_rayleigh
    """
    if 'density' in kargs.keys():
        del kargs['density']
    hist, bin_edges = np.histogram(x, density=False, **kargs)
    cum_hist = np.cumsum(hist)/x.shape[0]
    theo = cdf(bin_edges[1:], mean=mean, std=std)
    return np.max(np.abs(theo-cum_hist))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Décorateur timeout
^^^^^^^^^^^^^^^^^^

Created on Mon Nov 23 10:47:50 2020

@author: Cyrile Delestre
"""

import signal
from functools import wraps

def _handler_timeout(signum, frame):
    raise TimeoutError

def timeout(time: int, *args, **kwargs):
    r"""
    Décorateur simple d'utilisation permettant l'implémentation d'un timeout 
    sur une fonction, timeout exprimé en en milli-seconde.
    
    Parameters
    ----------
    time : int
        temps en milli-seconde
    
    Examples
    --------
    >>> import time
    >>> from dstk.utils import timeout
    
    >>> @time_out(100)
    >>> def fun(time_wait: int=10):
    >>>     time.sleep(time_wait/1000)
    >>>     print("ok!")
    
    >>> fun(10)
    ok!
    >>> fun(150)
    TimeoutError: Timeout de la fonction fun déclanchée après 100 ms !
    """
    def decorator(fun):
        @wraps(fun)
        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handler_timeout)
            signal.setitimer(signal.ITIMER_REAL, time/1000)
            try:
                res = fun(*args, **kwargs)
            except TimeoutError:
                signal.alarm(0)
                raise TimeoutError(
                    f"Timeout de la fonction {fun.__name__} déclanchée "
                    f"après {time} ms !"
                )
            signal.alarm(0)
            return res
        return wrapper
    return decorator



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Méthodes utiles pour l'exploitation PyTorch

Created on Thu Oct 31 09:42:49 2019

@author: Cyrile Delestre
"""
from time import time
from collections import defaultdict
from functools import partial
from typing import (Union, Optional, Dict, List, Callable, NewType, Any,
                    Iterable, Tuple)

import numpy as np
from scipy.stats import rankdata

from pandas.core.frame import DataFrame

from sklearn import __version__ as sklearn_version
from sklearn.model_selection import ParameterSampler
from sklearn.model_selection._validation import _fit_and_score
from sklearn.base import clone
from joblib import Parallel, delayed

from dstk.utils import list_of_dict_2_dict_of_list as l2d
from ._base import BaseEnvironnement

PyTorchEstimators = NewType('PyTorchEstimators', BaseEnvironnement)

class RandomizedSearchOnline:
    r"""
    Classe de random search des hyper-paramètres dans un contexte d'algorithme 
    online. Le fonctionnement de RandomizedSearchOnline est très semblable au
    fonctionnement de RandomizedSearchCV de Scikit-Learn (pour plus de 
    détails sur les arguments, se référer à la doc Scikit-Learn 
    `RandomizedSearchCV`_).
    
    .. _RandomizedSearchCV: https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html
    
    Parameters
    ----------
    estimator : PyTorchEstimators
        Un objet de type estimateur online. Les méthodes indispensables sont :
           - fit: (X, y=None)
               Fonction d'entrainement
           - fit_online: (X, y=None, predict=False)
               Fonction d'une phase online (entrainement + inférence)
    param_distributions : Dict[str, Any]
        Dictionnaire avec pour clefs les paramètres qu'on souhaite faire 
        varier (string). Pour chaque clef, une liste de l'univers possible 
        doit être décrite pour chaque paramètre.
    n_iter : int
        Nombre d'explorations distinctes dans l'univers des hyper-paramètres 
        de param_distributions (par défaut 10)
    scoring : Optional[Union[str, Callable, List[str], Tuple[str], Dict[str, Callable]]]
        None, string, list/tuple ou dict. Simple ou plusieurs fonction(s) de 
        scoring. Dans le cadre online, utiliser les fonctions de scoring 
        présentes dans utils.metrics. S'il s'agit d'une metrique custom, elle 
        doit avoir la forme :
            - my_metric_fun(estimator, X, y_true) 
                -> y_estim = estimator.fit_online(X, y_true, predict=True) 
                -> return metric_fun(y_estim, y_true)
        Par défaut, le classement se fait par ordre décroissant. Si fonction
        coût, prévoir une inversion du score dans la fonction de scoring.
    n_jobs : Optional[int]
        Nombre de jobs en parallèle. Si None, prendra 1 thread dans un 
        contexte joblib.parallel_backend. Si -1, prendra tous les threads du 
        processeur. Attention ! Si processeur multi-threadé, perte de 
        performances et prendre num_thread/2 (par défaut None).
    pre_dispatch : Optional[str, int]
        Contrôle le nombre de jobs devant être dispatchés durant l'éxécution. 
        Réduire le nombre permet d'éviter l'explosion de mémoire RAM (par 
        défaut '2*n_jobs').
    refit : Union[bool, str, Callable]
        Permet de réaliser le refit avec les paramètres ayant réalisé le 
        meilleur score. Pour de l'évaluation multi-metric, besoin d'un string 
        pour déterminer le choix du critère sur lequel sera fait le choix des 
        hyper-paramètres. refit peut également être une fonction qui retourne 
        ``best_index_`` pour un ``results_``. ``best_index_``, ``best_score_``,
        ``best_params_`` et ``best_estimator_`` seront des arguments existant 
        seulement s'il y a eu refit. (par défaut True)
    monte_carlo : int
        permet d'exécuter plusieur fois une topologie pour avec des scores 
        moyennés (par défaut 1, pas de Monte-Carlo).
    verbose : int
        Contrôle le niveau de verbosité (plus ce nombre est grand, plus des 
        informations seront marquées, par défaut 0).
    random_state : Optional[int]
        seed du ParameterSampler.
    error_score : str
        'raise' ou numeric. Comportement en cas d'erreur sur le score : 
        'raise' renvoit une erreur, sinon remplace par la valeur indiquée le 
        résultat du score (par exemple np.nan, par défaut 'raise').
    return_train_score : bool
        retourne le ou les scores sur le dataset de train (en plus de l'éval). 
        Cependant, le classement des scores se fera toujours sur le dataset 
        d'éval.
    
    Notes
    -----
    Si l'estimateur entré dans l'arguement estimator ne possède pas les 
    méthodes fit et fit_online, retourne une erreur.
    """
    def __init__(self,
        estimator: PyTorchEstimators,
        param_distributions: Dict[str, Any],
        n_iter: int=10,
        scoring: Optional[Union[str, Callable, List[str], Tuple[str], Dict[str, Callable]]]=None,
        n_jobs: Optional[int]=None,
        refit: Union[bool, str, Callable]=True,
        monte_carlo: int=1,
        verbose: int=0,
        pre_dispatch: Optional[Union[str, int]]='2*n_jobs',
        random_state: Optional[int]=None,
        error_score: str='raise',
        return_train_score: bool=False
    ):
        self.estimator = estimator
        self.param_distributions = param_distributions
        self.n_iter = n_iter
        self.scoring = scoring
        self.n_jobs = n_jobs
        self.refit = refit
        self.monte_carlo = monte_carlo
        self.verbose = verbose
        self.pre_dispatch = pre_dispatch
        self.random_state = random_state
        self.error_score = error_score
        self.return_train_score = return_train_score
        
        if not hasattr(estimator, 'fit'):
            raise TypeError("La fonction fit n'est pas implémentée à "
                            "l'estimator.")
        if not hasattr(estimator, 'fit_online'):
            raise TypeError(
                f"La classe {self.__name__} est destinée à des modèles Online "
                "qui possèdent comme méthode fit_online. Cette dernière n'est "
                "pas implémentée, vérifier qu'il s'agit bien d'une approche "
                "Online et voir si l'héritage BaseRegessorOnline ou "
                "BaseClassifierOnline a bien été hérité."
            )
        if not isinstance(monte_carlo, int) and monte_carlo < 1:
            raise TypeError(
                "La variable monte_carlo doit être un entier int supérieur "
                "ou égale à 1."
            )

    def fit(
        self,
        X: Iterable,
        y: Optional[Iterable]=None,
        idx_stop_train: Optional[int]=None,
        gap: int=0,
        idx_skip: Optional[Union[np.ndarray, List[int]]]=None,
        **fit_params
    ):
        r"""
        Applique le fit de l'estimateur sur l'univers des hyper-paramètres.
        
        Parameters
        ----------
        X : Iterable
            array-like ou Dataset de PyTorch. Dataset d'entrainement. Les 
            observations doivent rester dans l'ordre de dépendance.
        y : Iterable
            array-like [n-samples] ou [n-samples, n-output], optionnel. Target 
            relative à X. Si None, apprentissage non supervisé.
        idx_stop_train : Optional[int]
            Indice de l'horizon d'apprentissage.
        gap : int
            Permet de définir un gap entre l'apprentissage du modèle et la 
            prédiction. Par défaut 0, aucun gap. Dans la pratique le gap est 
            à minima égal à l'horizon de prédiction.
        idx_skip : Optional[Union[np.ndarray, List[int]]]
            array-like d'indexe que l'on souhaite extraire de l'évaluation 
            (par exemple les zones d'événements rares, etc.). Important : Si 
            on  dans le cadre d'un random search il faut soustraire 
            idx_stop_train aux indices à skiper.
        **fit_params :
            Paramètres passés à la fonction fit de l'estimateur.
        """
        if idx_stop_train is None or isinstance(idx_stop_train, (float, str)):
            raise ValueError(
                "La variable idx_stop_train est de type "
                f"{type(idx_stop_train).__name__}. Ce doit être un integer "
                "(int) représentant un indice."
            )
        if isinstance(X, (np.ndarray, DataFrame)):
            len_dataset = X.shape[0]
        else:
            len_dataset = len(X)
        if idx_stop_train >= len_dataset-1:
            raise ValueError(
                "L'indice idx_stop_train doit être strictement inférieur à "
                "la taille du dataset X."
            )

        fit_params_scorers = fit_params.copy()
        fit_params_scorers['build'] = False
        if callable(self.scoring):
            self.multimetric_ = False
            scorers = {
                'score': partial(
                    self.scoring,
                    gap = gap,
                    idx_skip = idx_skip,
                    kargs_fit = fit_params_scorers
                )
            }
        elif isinstance(self.scoring, dict):
            self.multimetric_ = True
            scorers = {
                kk: partial(
                    ss,
                    gap = gap,
                    idx_skip = idx_skip,
                    kargs_fit = fit_params_scorers
                ) for kk, ss in self.scoring.items()
            }
        elif isinstance(self.scoring, (list, tuple)):
            self.multimetric_ = True
            scorers = {
                ii: partial(
                    ii,
                    gap = gap,
                    idx_skip = idx_skip,
                    kargs_fit = fit_params_scorers
                ) for ii in self.scoring
            }
        else:
            raise ValueError(
                "Le paramètre scoring doit prendre soit une fonction de "
                "scoring soit un dictionnaire contenant des fonctions score "
                "ou une liste ou tuple de fonctions score. Pas "
                f"{self.scoring}."
            )

        if self.multimetric_:
            if self.refit is not False and (
                    not isinstance(self.refit, str) or
                    self.refit not in scorers) and not callable(self.refit):
                raise ValueError(
                    "Si multi-métric, il faut que le refit soit True, dans "
                    "ce cas la meilleur métrique sera utilisée. Ou sinon un "
                    "callable ou le nom correspondant à l'une des mesures de "
                    "performance contenues dans scoring. Et non "
                    f"refit={self.refit}."
                )
            else:
                refit_metric = self.refit
        else:
            refit_metric = 'score'

        train, test = (
            list(range(idx_stop_train)),
            list(range(idx_stop_train, len_dataset))
        )
        candidate_params = list(
            ParameterSampler(
                param_distributions = self.param_distributions,
                n_iter = self.n_iter,
                random_state = self.random_state
            )
        )
        n_candidates = len(candidate_params)

        print(
            f"Fitting {n_candidates} candidates with {self.monte_carlo} "
            f"Monte-Carlo iterations. {n_candidates*self.monte_carlo} "
            "models should be fitted.",
            flush=True
        )

        estimator = clone(self.estimator)
        all_candidate_params = []
        all_out = []
        with Parallel(n_jobs = self.n_jobs,
                      verbose = self.verbose,
                      pre_dispatch = self.pre_dispatch) as parallel:
            # La fonction _fit_and_score retourne dans l'ordre :
            # [train_scores, test_scores, fit_time, score_time]
            out = parallel(
                delayed(_fit_and_score)(
                    estimator = clone(estimator),
                    X = X,
                    y = y,
                    train = train,
                    test = test,
                    scorer = scorers,
                    verbose = self.verbose,
                    parameters = parameters,
                    fit_params = fit_params,
                    return_train_score = self.return_train_score,
                    return_n_test_samples = True,
                    return_parameters = False,
                    return_times = True,
                    return_estimator = False,
                    error_score = self.error_score
                )
                for parameters in candidate_params 
                for mc in range(self.monte_carlo)
            )
        if len(out) == 0:
            raise ValueError(
                "La fonction fit ne s'est pas bien déroulée. Le problème "
                "doit venir de param_distributions. Est-il non-vide ?"
            )
        if len(out) != n_candidates*self.monte_carlo:
            raise ValueError(
                "Le nombre de résultats en sortie de la validation différe "
                "du nombre de tests souhaité. Il y a eu {} exécution(s) alors "
                "que {} étai(en)t attendue(s)."
                .format(len(out), n_candidates*self.monte_carlo)
            )
        all_candidate_params.extend(candidate_params)
        all_out.extend(out)

        results = self._format_results(all_candidate_params, scorers, all_out)

        if self.refit or not self.multimetric_:
            if callable(self.refit):
                self.best_index_ = self.refit(results)
                if not isinstance(self.best_index_, (int, np.integer)):
                    raise TypeError(
                        "best_index_ n'est pas un integer."
                    )
                if (self.best_index_ < 0 or 
                        self.best_index_ >= len(results["params"])):
                    raise IndexError(
                        "``best_index_`` est out of range, doit être compris "
                        "entre 0 et {} et vaut {}."
                        .format(len(results["params"]), self.best_index_)
                    )
            else:
                self.best_index_ = (
                    results[f"rank_test_{refit_metric}"].argmin()
                )
                self.best_score_ = (
                    results[(
                        f"mean_test_{refit_metric}" if self.monte_carlo > 1
                        else f"test_{refit_metric}"
                    )][self.best_index_]
                )
            self.best_params_ = results["params"][self.best_index_]

        if self.refit:
            self.best_estimator_ = clone(estimator)\
                .set_params(**self.best_params_)
            refit_start_time = time()
            if y is not None:
                self.best_estimator_.fit(X, y, **fit_params)
            else:
                self.best_estimator_.fit(X, **fit_params)
            refit_end_time = time()
            self.refit_time_ = refit_end_time - refit_start_time

        self.scorer_ = scorers if self.multimetric_ else scorers['score']

        self.results_ = results

        return self

    def _format_results(
        self,
        candidate_params: List[Dict[str, Any]],
        scorers: Union[Callable, Dict[str, Callable]],
        out: List[float]
    ):
        r"""\
        Fontion de formatage des résultats interne à la classe.
        
        Parameters
        ----------
        candidate_params : List[Dict[str, Any]]
            liste des dictionnaires générée par ParameterSampler.
        scorers : Union[Callable, Dict[str, Callable]]
            dictionnaire ou callable des scores.
        out : List[float]
            sortie des apprentissages avec en sortie :
                - return_train_score = True
                    [train_scores, test_scores, fit_time, score_time]
                - return_train_score = False
                    [test_scores, fit_time, score_time]
        
        Returns
        -------
        results : Dict[str, Any]
            dictionnaire contenant les paramètres avec le rang associé en 
            fonction des scores.
        
        Notes
        -----
        Cette méthode est inspiré de la méthode _format_results qu'on 
        peut trouver dans Scikit-Learn à quelques adaptations près.
        """
        n_candidates = len(candidate_params)
        if sklearn_version < '0.20':
            if self.return_train_score:
                (train_score_dicts, test_score_dicts,
                 test_sample_counts, fit_time, score_time) = zip(*out)
            else:
                (test_score_dicts, test_sample_counts, fit_time,
                 score_time) = zip(*out)
        else:
            out_ = l2d(out)
            fit_failed = out_['fit_failed']
            test_score_dicts = out_['test_scores']
            # test_sample_counts = out_['n_test_samples']
            fit_time = out_['fit_time']
            score_time = out_['score_time']
            if self.return_train_score:
                train_score_dicts = out_['train_scores']

        test_scores = {
            key: np.asarray(
                [score[key] for score in test_score_dicts]
            ) for key in test_score_dicts[0]
        }
        if self.return_train_score:
            train_scores = {
                key: np.asarray(
                    [score[key] for score in train_score_dicts]
                ) for key in train_score_dicts[0]
            }

        results = {}

        def _store(key_name, array, monte_carlo=False, rank=False):
            """Fonction interne d'aide au remplissage de results_"""
            if self.monte_carlo > 1:
                array = (
                    np.array(array, dtype=np.float64)
                    .reshape(n_candidates, self.monte_carlo)
                )
                if monte_carlo:
                    for mc_i in range(self.monte_carlo):
                        results[f"mc{mc_i}_{key_name}"] = array[:, mc_i]
                if key_name != 'fit_failed':
                    array_means = np.average(array, axis=1)
                    results[f"mean_{key_name}"] = array_means
                    array_stds = np.std(array, axis=1)
                    results[f"std_{key_name}"] = array_stds
                else:
                    array_means = np.average(array, axis=1)
                    results[f"ratio_{key_name}"] = array_means
                if rank:
                    results[f"rank_{key_name}"] = np.asarray(
                        rankdata(-array_means, method='min'),
                        dtype=np.int32
                    )
            elif self.monte_carlo == 1:
                array = np.array(array, dtype=np.float64).ravel()
                results[f"{key_name}"] = array
                if rank:
                    results[f"rank_{key_name}"] = np.asarray(
                        rankdata(-array, method='min'),
                        dtype=np.int32
                    )
            else:
                raise AttributeError(
                    "L'attribue monte_carlo doit être un entier supérieur "
                    f"ou égal à 1 et non égal à {self.monte_carlo}."
                )

        if sklearn_version >= '0.20':
            _store('fit_failed', fit_failed)
        _store('fit_time', fit_time)
        _store('score_time', score_time)

        # Utilise un MaskedArray pour masquez tous les endroits où les 
        # paramètres ne sont pas applicable pour ce candidat.
        # Utilise defaultdict car chaque candidat peut ne pas contenir tous 
        # les paramètres.
        param_results = defaultdict(
            partial(
                np.ma.MaskedArray,
                np.empty(n_candidates,),
                mask=True,
                dtype=object
            )
        )
        for cand_i, params in enumerate(candidate_params):
            for name, value in params.items():
                # Un masque entièrement vide est créé pour la première 
                # occurence de `name` `f"param_{name}"`.
                # La définition de la valeur à un index démasque également 
                # cet index.
                param_results[f"param_{name}"][cand_i] = value
        
        results.update(param_results)
        # Store la liste des paramètre
        results['params'] = candidate_params

        for scorer_name in scorers.keys():
            _store(
                key_name = f"test_{scorer_name}",
                array = test_scores[scorer_name],
                monte_carlo = True,
                rank = True
            )
            if self.return_train_score:
                _store(
                    key_name = f"train_{scorer_name}",
                    array = train_scores[scorer_name],
                    monte_carlo = True
                )

        return results

    def _check_refit(self):
        r"""
        Méthodes privées de check refit.
        """
        if not hasattr(self, 'best_estimator_'):
            raise ValueError(
                "Le modèle best_estimator_ n'existe pas. Il ce peut que "
                "vous n'ayez pas utilisé l'option refit de "
                "RandomizedSearchOnline ou vous n'avez pas encore exécuté "
                "le fit."
            )

    def predict(self, X: Iterable):
        r"""
        Méthode permettant de wrapper le predict de l'estimateur.
        
        Parameters
        ----------
        X : Iterable
            observations
        """
        self._check_refit()
        return self.best_estimator_.predict(X)

    def predict_proba(self, X: Iterable):
        r"""
        Méthode permettant de wrapper le predict_broba de l'estimateur pour 
        un classifier.
        
        Parameters
        ----------
        X : Iterable
            observations
        """
        self._check_refit()
        if self.estimator._estimator_type != 'classifier':
            raise ValueError(
                "L'estimateur n'est pas un classifier, il s'agit d'un {}. "
                "Il n'a donc pas de méthode predict_proba."
                .format(self.estimator._estimator_type)
            )
        return self.best_estimator_.predict_proba(X)

    def fit_online(self, **kargs):
        r"""
        Méthode permettant de wrapper le fit_online.
        
        Parameters
        ----------
        kargs :
            arguments de fit_online de l'estimateur.
        """
        self._check_refit()
        return self.best_estimator_.fit_online(**kargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classes Handler pour les différents type de connexion (SQLite,SQLAlchemy, 
Warp10, Artimon, etc.).

Created on Sat Apr 11 06:41:28 2020

@author: Cyrile Delestre
"""

import os
import json
import logging
import sqlite3
from typing import List, Optional
from datetime import datetime as dt

# Mapping des champs possibles d'un logging avec les type SQLite3
MAP_TYPE_HANDLER = dict(
    asctime = 'TEXT',
    created = 'REAL',
    filename = 'TEXT',
    funcName = 'TEXT',
    levelname = 'TEXT',
    levelno = 'TEXT',
    lineno = 'INTEGER',
    message = 'TEXT',
    module = 'TEXT',
    msecs = 'INTEGER',
    name = 'TEXT',
    pathname = 'TEXT',
    process = 'INTEGER',
    processName = 'TEXT',
    relativeCreated = 'REAL',
    thread = 'INTEGER',
    threadName = 'TEXT',
    engine = 'TEXT'
)

class HandlerBase(logging.Handler):
    r"""
    Classe de base permettant de créer un handler à partir d'une liste de 
    Handler.
    
    Parameters
    ----------
    format : List[str]
        liste des attributs du LogRecord qu'on souhaite récupérer et 
        stocker. Liste des attributs de LogRecord présentés à l'URL :
            https://docs.python.org/3/library/logging.html#logrecord-attributes
        asctime : str
            Timestamps de création du log interprétable 
            humainement ;
        created : float
            Timestamps de création du log ;
        filename : str
            Nom du fichier dans lequel a été appelé le logging ;
        funcName : str
            Nom de la fonction contenant l'appel du logging ;
        levelname : str
            Nom de level d'alerte ('DEBUG', 'INFO', 'WARNING', 'ERROR', 
            'CRITICAL') ;
        levelno : int
            Numérique associé au level d'alerte ;
        message : str
            Message envoyé par le logging (à définir par l'utilisateur) ;
        module : str
            Nom du module ;
        msecs : int
            Temps de création de LogRicord (en milliseconde) ;
        name : str
            Nom du logger (permet d'identifier les différents logger) ;
        pathname : str
            Chemin complet du fichier dans lequel a été appelé le logging (si 
            disponible);
        process : int
            Process ID (si disponible) ;
        processName : str
            Nom du process ID (si disponible) ;
        relativeCreated : float
            Temps en millisecondes de création de LogRecord, par rapport au 
            temps de chargement du module de logging ;
        thread : int
            Thread ID (si disponible) ;
        threadName : str
            Nom du thread ID (si disponible).
        La suite est composée des attributs de LogRecord créés
            engine : str
                Nom du moteur ('sqlite', 'sqlalchemy', Warp10, etc.).
    
    See also
    --------
    SQLiteHandler, SQLAlchemyHandler
    """
    def __init__(
        self,
        format: List[str]=['name', 'message']
    ) -> None:
        logging.Handler.__init__(self)
        self.format = format

    def _asctime(self, record: logging.LogRecord) -> None:
        r"""
        Méthode ajoutant au LogRecord l'attribut asctime qui est le timestamp
        (created) du log au format lisible.
        """
        record.asctime = (
            dt
            .fromtimestamp(record.created)
            .isoformat()
        )

    def _message(self, record: logging.LogRecord) -> None:
        r"""
        Méthode ajoutant au LogRecord l'attribut message qui est une copie de 
        l'attribut msg au format json encodé en UTF-8.
        """
        record.message = json.dumps(record.msg).encode('utf-8')

    def _engine(self, record: logging.LogRecord, engine: str) -> None:
        r"""
        Méthode ajoutant au LogRecord l'attribut engine.
        """
        record.engine = engine

    def _extract_format(self, record: logging.LogRecord) -> List[str]:
        r"""
        Méthode extrayant les informations du LogRecord sélectionnées par 
        l'attribut format.
        """
        rec = record.__dict__
        return list(map(lambda x: rec[x], self.format))


    def emit(self, record: logging.LogRecord) -> None:
        r"""
        Méthode d'envoi des logs à implémenter.
        """
        raise NotImplementedError(
            "Il faut implémenter la méthode emit(self, record) à la classe "
            f"{self.__name__} qui hérite de la classe HandlerBase."
        )

class SQLiteHandler(HandlerBase):
    r"""
    Classe permettant de créer un handler à partir d'une liste de Handler 
    vers une base SQLite.
    
    Parameters
    ----------
    path_sql : str
        chemin vers la base SQLite des logs.
    format : List[str]
        liste des attributs du LogRecord qu'on souhaite récupérer et stocker. 
        Liste des attributs de LogRecord présentés à l'URL :
            https://docs.python.org/3/library/logging.html#logrecord-attributes
        asctime : str
            Timestamps de création du log interprétable humainement ;
        created : float
            Timestamps de création du log ;
        filename : str
            Nom du fichier dans lequel a été appelé le logging ;
        funcName : str
            Nom de la fonction contenant l'appel du logging ;
        levelname : str
            Nom de level d'alerte ('DEBUG', 'INFO', 'WARNING', 'ERROR', 
            'CRITICAL') ;
        levelno : int
            Numérique associé au level d'alerte ;
        message : str
            Message envoyé par le logging (à définir par l'utilisateur) ;
        module : str
            Nom du module ;
        msecs : int
            Temps de création de LogRicord (en milliseconde) ;
        name : str
            Nom du logger (permet d'identifier les différents logger) ;
        pathname : str
            Chemin complet du fichier dans lequel a été appelé le logging (si 
            disponible);
        process : int
            Process ID (si disponible) ;
        processName : str
            Nom du process ID (si disponible) ;
        relativeCreated : float
            Temps en millisecondes de création de LogRecord, par rapport au 
            temps de chargement du module de logging ;
        thread : int
            Thread ID (si disponible) ;
        threadName : str
            Nom du thread ID (si disponible).
        La suite est composée des attributs de LogRecord créés
            engine : str
                Nom du moteur ('sqlite', 'sqlalchemy', etc.).
    base_name : str
        Nom de la base (par dégaut "logs").
    :**kargs_sql: arguments d'entrée au connecteur 
        SQLite3 :
            sqlite3.connect(path_sql, **kargs_sql)
    
    Examples
    --------
    Nous allons utiliser timer pour extraire les informations temporelles 
    d'éxécution d'une fonction.
    
    >>> import logging
    >>> from time import sleep
    >>> from dstk.metrics import SQLiteHandler, timer
    >>> 
    >>> # Création du logger
    >>> logging.basicConfig(
    >>>     level = logging.INFO,
    >>>     handlers = [
    >>>         SQLiteHandler(
    >>>             path_sql = 'metrics.log',
    >>>             format = ['asctime', 'name', 'message', 'threadName']
    >>>         )
    >>>     ]
    >>> )
    >>> logger = logging.getLogger('python.metrics.timer')
    
    >>> @timer(logger=logger, delta_time=True)
    >>> def function_test(a=1, b=1, sleep_time=0.5):
    >>>     r"Docstring de la fonction fonction_test"
    >>>     sleep(sleep_time)
    >>>     return a+b
    >>> 
    >>> function_test()
    2
    >>> function_test(2, 2)
    4
    
    Les logs sont stockés dans la base SQLite "metrics.log".
    
    >>> import sqlite3
    >>> from dstk.data import PandasSQL
    >>> 
    >>> log = PandasSQL(sqlite3.connect("metrics.log"))
    >>> 
    >>> log("SELECT asctime, name, threadName FROM logs;")
                          asctime                  name  threadName
    0  2020-04-13T13:31:06.973864  python.metrics.timer  MainThread
    1  2020-04-13T13:31:11.469728  python.metrics.timer  MainThread
    >>> 
    >>> log("SELECT message FROM logs;")
                                                 message
    0  {"name": "function_test", "dt": "0:00:00.500830"}
    1  {"name": "function_test", "dt": "0:00:00.500830"}
    
    See also
    --------
    SQLAlchemyHandler, timer, timer_class
    """
    def __init__(
        self,
        path_sql: str,
        format: List[str]=['name', 'message'],
        base_name: str='logs',
        **kargs_sql
    ) -> None:
        if 'msg' in format:
            raise AttributeError(
                "Ne pas utiliser l'attribut 'msg' qui est technique, mais "
                "utiliser plutôt l'attribut 'message'."
            )
        super().__init__(format = format)
        self.kargs_sql = kargs_sql
        self.path_sql = path_sql
        self.base_name = base_name
        conn = sqlite3.connect(path_sql, **kargs_sql)
        conn.execute(
            f"CREATE TABLE IF NOT EXISTS {self.base_name}("
                f"{', '.join([f'{cc} {MAP_TYPE_HANDLER[cc]}' for cc in self.format])}"
            ")"
        )
        conn.commit()
        conn.close()
        del conn

    def emit(self, record: logging.LogRecord) -> None:
        self._asctime(record)
        self._message(record)
        self._engine(record, 'sqlite')
        log_in = self._extract_format(record)
        conn = sqlite3.connect(self.path_sql, **self.kargs_sql)
        conn.execute(
            f"INSERT INTO {self.base_name}({', '.join(self.format)}) "
            f"VALUES({', '.join(len(self.format)*'?')})",
            log_in
        )
        conn.commit()
        conn.close()
        del conn

class SQLAlchemyHandler(HandlerBase):
    r"""
    Classe permettant de créer un handler à partir d'une liste de Handler 
    vers une base interfacée à SQLAlchemy.
    
    Parameters
    ----------
    path_sql : str
        URL de connexion SQLAlchemy. L'URL de connexion est définie par :
            "dialect+driver://usernam:password@host:port/database"
    format : List[str]
        liste des attributs du LogRecord qu'on souhaite récupérer et stocker. 
        Liste des attributs de LogRecord présentés à l'URL :
            https://docs.python.org/3/library/logging.html#logrecord-attributes
        asctime : str
            Timestamps de création du log interprétable humainement ;
        created : float
            Timestamps de création du log ;
        filename : str
            Nom du fichier dans lequel a été appelé le logging ;
        funcName : str
            Nom de la fonction contenant l'appel du logging ;
        levelname : str
            Nom de level d'alerte ('DEBUG', 'INFO', 'WARNING', 'ERROR', 
            'CRITICAL') ;
        levelno : int
            Numérique associé au level d'alerte ;
        message : str
            Message envoyé par le logging (à définir par l'utilisateur) ;
        module : str
            Nom du module ;
        msecs : int
            Temps de création de LogRicord (en milliseconde) ;
        name : str
            Nom du logger (permet d'identifier les différents logger) ;
        pathname : str
            Chemin complet du fichier dans lequel a été appelé le logging (si 
            disponible);
        process : int
            Process ID (si disponible) ;
        processName : str
            Nom du process ID (si disponible) ;
        relativeCreated : float
            Temps en millisecondes de création de LogRecord, par rapport au 
            temps de chargement du module de logging ;
        thread : int
            Thread ID (si disponible) ;
        threadName : str
            Nom du thread ID (si disponible).
        La suite est composée des attributs de LogRecord créés
            engine : str
                Nom du moteur ('sqlite', 'sqlalchemy', etc.).
    base_name : str
        Nom de la base (par dégaut "logs").
    **kargs_sql :
        arguments d'entrée au connecteur SQLAlchemy :
            sqlalchemy.create_engine(path_sql, **kargs_sql)
    
    Notes
    -----
    Cette classe nécessite SQLAlchemy, si ce package n'est pas présent dans 
    votre environnement il suffit d'ajouter le package "sqlalchemy" au fichier 
    environment.yml et de lancer le script bash bash_install_env.sh dans le 
    répertoire shell du projet.
    
    Examples
    --------
    Nous allons utiliser timer pour extraire les informations temporelles 
    d'éxécution d'une fonction et les stocker dans une table SQLite 
    interfacée par SQLAlchemy.
    
    >>> import logging
    >>> from time import sleep
    >>> from dstk.metrics import SQLAlchemyHandler, timer
    >>> 
    >>> # Création du logger
    >>> logging.basicConfig(
    >>>     level = logging.INFO,
    >>>     handlers = [
    >>>         SQLAlchemyHandler(
    >>>             path_sql = 'sqlite:///metrics.log',
    >>>             format = ['asctime', 'name', 'message', 'threadName']
    >>>         )
    >>>     ]
    >>> )
    >>> logger = logging.getLogger('python.metrics.timer')
    
    >>> @timer(logger=logger, delta_time=True)
    >>> def function_test(a=1, b=1, sleep_time=0.5):
    >>>     r"Docstring de la fonction fonction_test"
    >>>     sleep(sleep_time)
    >>>     return a+b
    >>> 
    >>> function_test()
    2
    >>> function_test(2, 2)
    4
    
    Les logs sont stockés dans la base SQLite "metrics.log".
    
    >>> import sqlite3
    >>> import dstk.data import PandasSQL
    >>> 
    >>> log = PandasSQL(sqlite3.connect("metrics.log"))
    >>>
    >>> log("SELECT asctime, name, threadName FROM logs;")
                          asctime                  name  threadName
    0  2020-04-13T13:31:06.973864  python.metrics.timer  MainThread
    1  2020-04-13T13:31:11.469728  python.metrics.timer  MainThread
    >>> 
    >>> log("SELECT message FROM logs;")
                                                 message
    0  {"name": "function_test", "dt": "0:00:00.500830"}
    1  {"name": "function_test", "dt": "0:00:00.500830"}
    
    See also
    --------
    SQLiteHandler, timer, timer_class
    """
    def __init__(
        self,
        path_sql: str,
        format: List[str]=['name', 'message'],
        base_name: str='logs',
        **kargs_sql
    ) -> None:
        try:
            global sqlalchemy
            import sqlalchemy
            from sqlalchemy.types import Integer, Float, String
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Le module SQLAlchemy n'est pas installé : l'instruction "
                "import sqlalchemy renvoit une erreur "
                "ModuleNotFoundError. Rajouter sqlalchemy dans les "
                "packages à installer dans le fichier environment.yml du "
                "projet."
            )
        if 'msg' in format:
            raise AttributeError(
                "Ne pas utiliser l'attribut 'msg' qui est technique, mais "
                "utiliser plutôt l'attribut 'message'."
            )
        super().__init__(format = format)
        self.path_sql = path_sql
        self.kargs_sql = kargs_sql
        self.base_name = base_name
        map_type = dict(
            TEXT = String(258),
            INTEGER = Integer(),
            REAL = Float()
        )
        engine = sqlalchemy.create_engine(path_sql, **kargs_sql)
        with engine.connect() as conn:
            conn.execute(
                f"CREATE TABLE IF NOT EXISTS {self.base_name}("
                    f"{', '.join([f'{cc} {map_type[MAP_TYPE_HANDLER[cc]]}' for cc in self.format])}"
                ")"
            )
        del engine

    def emit(self, record: logging.LogRecord) -> None:
        self._asctime(record)
        self._message(record)
        self._engine(record, 'sqlalchemy')
        log_in = self._extract_format(record)
        engine = sqlalchemy.create_engine(self.path_sql, **self.kargs_sql)
        with engine.connect() as conn:
            conn.execute(
                f"INSERT INTO {self.base_name}({', '.join(self.format)}) "
                f"VALUES({', '.join(len(self.format)*'?')})",
                log_in
            )
        del engine

class BaseArtimon(HandlerBase):
    r"""
    Classe de base permettant de gérer les données à envoyer dans Warp10 
    via Artimon. Cette classe n'est pas à utiliser directement, elle est 
    utiliser en tant que classe mère aux classes ArtimonAPIHandler et 
    ArtimonLocalHandler.
    
    Parameters
    ----------
    producer : str
        Nom du producer vers lequel les métriques seront envoyées.
    format : List[str]
        liste des attributs du LogRecord qu'on souhaite récupérer et 
        stocker. Liste des attributs de LogRecord présentés à l'URL :
            https://docs.python.org/3/library/logging.html#logrecord-attributes
        asctime : str
            Timestamps de création du log interprétable humainement ;
        created : float
            Timestamps de création du log ;
        filename : str
            Nom du fichier dans lequel a été appelé le logging ;
        funcName : str
            Nom de la fonction contenant l'appel du logging ;
        levelname : str
            Nom de level d'alerte ('DEBUG', 'INFO', 'WARNING', 'ERROR', 
            'CRITICAL') ;
        levelno : int
            Numérique associé au level d'alerte ;
        message : str
            Message envoyé par le logging (à définir par l'utilisateur) ;
        module : str
            Nom du module ;
        msecs : int
            Temps de création de LogRicord (en milliseconde) ;
        name : str
            Nom du logger (permet d'identifier les différents logger) ;
        pathname : str
            Chemin complet du fichier dans lequel a été appelé le logging (si 
            disponible);
        process : int
            Process ID (si disponible) ;
        processName : str
            Nom du process ID (si disponible) ;
        relativeCreated : float
            Temps en millisecondes de création de LogRecord, par rapport au 
            temps de chargement du module de logging ;
        thread : int
            Thread ID (si disponible) ;
        threadName : str
            Nom du thread ID (si disponible).
        La suite est composée des attributs de LogRecord créés
            engine : str
                Nom du moteur ('sqlite', 'sqlalchemy', 'Warp10', etc.).
    hostname : Optional[str]
        nom facultatif du hostname (si non renseigné sera complété 
        automatiquement par Artimon).
    environment : Optional[str]
        nom facultatif de l'environnement (si non renseigné sera complété 
        automatiquement par Artimon).
    
    See also
    --------
    ArtimonAPIHandler, ArtimonLocalHendleur
    """
    def __init__(
        self, producer: str,
        format: List[str]=['name', 'msg'],
        hostname: Optional[str]=None,
        environment: Optional[str]=None
    ) -> None:
        super().__init__(format = format)
        self.producer = producer
        self.hostname = hostname
        self.environment = environment

    def parser(self, record: logging.LogRecord) -> str:
        name = record.name
        format = self.format
        if 'created' in self.format:
            format.remove('created')
        if 'name' in self.format:
            format.remove('name')
        if 'msg' in self.format:
            format.remove('msg')
        log_att = record.__dict__
        log_att = list(map(lambda x: log_att[x], format))
        tags = {kk: vv for kk, vv in zip(format, log_att)}
        if self.hostname is not None:
            tags['hostname'] = self.hostname
        if self.environment is not None:
            tags['environment'] = self.environnement
        message = [
            dict(
                i = dict(p=self.producer),
                m = dict(
                    n = name,
                    t = tags
                ),
                v = record.msg,
                c = dict(t=int(round(record.created*1000)))
            )
        ]
        return json.dumps(message)

class ArtimonAPIHandler(BaseArtimon):
    r"""
    Classe permettant de créer un handler designné pour Artimon le parser 
    Warp10 du SI Arkéa. Elle utilise le protocal POST pour envoyer la donnée 
    dans Warp10.
    
    Parameters
    ----------
    producer : str
        nom du producer Artimon/Warp10.
    host : str
        host de l'API Artimon (par défaut '127.0.0.1').
    port : str
        port de l'API Artimon (par défaut '9023').
    format : List[str]
        liste des attributs du LogRecord qu'on souhaite récupérer et stocker. 
        Liste des attributs de LogRecord présentés à l'URL :
            https://docs.python.org/3/library/logging.html#logrecord-attributes
        asctime : str
            Timestamps de création du log interprétable humainement ;
        created : float
            Timestamps de création du log ;
        filename : str
            Nom du fichier dans lequel a été appelé le logging ;
        funcName : str
            Nom de la fonction contenant l'appel du logging ;
        levelname : str
            Nom de level d'alerte ('DEBUG', 'INFO', 'WARNING', 'ERROR', 
            'CRITICAL') ;
        levelno : int
            Numérique associé au level d'alerte ;
        message : str
            Message envoyé par le logging (à définir par l'utilisateur) ;
        module : str
            Nom du module ;
        msecs : int
            Temps de création de LogRicord (en milliseconde) ;
        name : str
            Nom du logger (permet d'identifier les différents logger) ;
        pathname : str
            Chemin complet du fichier dans lequel a été appelé le logging (si 
            disponible);
        process : int
            Process ID (si disponible) ;
        processName : str
            Nom du process ID (si disponible) ;
        relativeCreated : float
            Temps en millisecondes de création de LogRecord, par rapport au 
            temps de chargement du module de logging ;
        thread : int
            Thread ID (si disponible) ;
        threadName : str
            Nom du thread ID (si disponible).
        La suite est composée des attributs de LogRecord créés
            engine : str
                Nom du moteur ('sqlite', 'sqlalchemy', 'Warp10', etc.).
    hostname : Optional[str]
        nom facultatif du hostname (si non renseigné sera complété 
        automatiquement par Artimon).
    environment : Optional[str]
        nom facultatif de l'environnement (si non renseigné sera complété 
        automatiquement par Artimon).
    
    Notes
    -----
    Avec cette classe il n'est pas possible d'utiliser le format "message", 
    il est important de ce référer à "msg" pour le message. De même il n'est 
    pas nécessaire de mettre "msg" dans l'argument format car le contenu du 
    message sera utilisé automatiquement comme argument de la série.
    
    Examples
    --------
    Envoi d'un timer mesurant de temps d'exécution d'une fonction et de sans 
    sortie.
    
    >>> import logging
    >>> from time import sleep
    >>> import numpy as np
    >>> from dstk.metrics import ArtimonAPIHandler, timer, sniffer
    >>> 
    >>> # Création du logger
    >>> logging.basicConfig(
    >>>     level = logging.INFO,
    >>>     handlers = [
    >>>         ArtimonAPIHandler(
    >>>             producer = 'NAME.producer',
    >>>             format = ['asctime', 'threadName']
    >>>         )
    >>>     ]
    >>> )
    >>> logger_timer = logging.getLogger('metrics.python.timer')
    >>> logger_sniffer = logging.getLogger('metrics.python.sniffer')
    
    >>> @timer(logger=logger_timer, delta_time=True)
    >>> @sniffer(logger=logger_sniffer)
    >>> def function_test(a=1, b=1, sleep_time=0.5):
    >>>     r"Docstring de la fonction fonction_test"
    >>>     sleep(sleep_time)
    >>>     return a+b
    
    >>> res = [
    >>>     function_test(a=ii, sleep_time=5*np.random.rand())
    >>>     for ii in range(10)
    >>> ]
    >>> res
    [1, 3, 4, 5, 6, 7, 8, 9, 10]
    
    Les données seront envoyés directement vers Warp10.
    
    See also
    --------
    ArtimonLocalHandler
    """
    def __init__(
        self,
        producer: str,
        host: str='127.0.0.1',
        port: str='9023',
        format: List[str]=['name', 'msg'],
        hostname: Optional[str]=None,
        environment: Optional[str]=None
    ) -> None:
        try:
            global requests
            import requests
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Le module requests n'est pas installé : l'instruction "
                "import requests renvoit une erreur "
                "ModuleNotFoundError. Rajouter requests dans les "
                "packages à installer dans le fichier environment.yml du "
                "projet."
            )
        if 'message' in format:
            raise AttributeError(
                "Ne pas utiliser l'attribut 'message' qui est technique, mais "
                "utiliser plutôt l'attribut 'msg'."
            )
        super().__init__(
            format = format,
            producer = producer,
            hostname = hostname,
            environment = environment
        )
        self.host = host
        self.port = port
        self.url = (
            f"http://{self.host}:{self.port}/metrics"
        )

    def emit(self, record: logging.LogRecord) -> None:
        self._asctime(record)
        self._engine(record, 'artimon')
        response = requests.post(
            self.url,
            data = self.parser(record)
        )
        if response.status_code != 200:
            stat = int(response.status_code/100)
            if stat == 4 or stat == 5:
                raise IOError(
                    "Le retour du puch vers Warp10 à renvoyer une erreur. L'API "
                    f"a renvoyer le code {response.status_code}. "
                    f"Message d'erreur :\n{response.text}"
                )
            else:
                Warning(
                    "Le retour du puch vers Warp10 n'est pas 200 mais "
                    f"{response.status_code}. Message de retour :\n"
                    f"{response.text}"
                )

class ArtimonLocalHandler(BaseArtimon):
    r"""
    Classe permettant de créer un handler designné pour Artimon le parser 
    Warp10 du SI Arkéa. Elle utilise le collecteur Artimon qui doit être 
    installé en local. Le collecteur Artimon va cherche les fichiers 
    finissant par l'extention .artimon dans le répertoire par défaut 
    '/var/run/artimonV2'.
    
    Parameters
    ----------
    producer : str
        nom du producer Artimon/Warp10.
    path : str
        répertoire du dépot des métriques pour le collecteur Artimon (par 
        défaut '/var/run/artimonV2).
    format : List[str]
        liste des attributs du LogRecord qu'on souhaite récupérer et stocker. 
        Liste des attributs de LogRecord présentés à l'URL :
            https://docs.python.org/3/library/logging.html#logrecord-attributes
        asctime : str
            Timestamps de création du log interprétable humainement ;
        created : float
            Timestamps de création du log ;
        filename : str
            Nom du fichier dans lequel a été appelé le logging ;
        funcName : str
            Nom de la fonction contenant l'appel du logging ;
        levelname : str
            Nom de level d'alerte ('DEBUG', 'INFO', 'WARNING', 'ERROR', 
            'CRITICAL') ;
        levelno : int
            Numérique associé au level d'alerte ;
        message : str
            Message envoyé par le logging (à définir par l'utilisateur) ;
        module : str
            Nom du module ;
        msecs : int
            Temps de création de LogRicord (en milliseconde) ;
        name : str
            Nom du logger (permet d'identifier les différents logger) ;
        pathname : str
            Chemin complet du fichier dans lequel a été appelé le logging (si 
            disponible);
        process : int
            Process ID (si disponible) ;
        processName : str
            Nom du process ID (si disponible) ;
        relativeCreated : float
            Temps en millisecondes de création de LogRecord, par rapport au 
            temps de chargement du module de logging ;
        thread : int
            Thread ID (si disponible) ;
        threadName : str
            Nom du thread ID (si disponible).
        La suite est composée des attributs de LogRecord créés
            engine : str
                Nom du moteur ('sqlite', 'sqlalchemy', 'Warp10', etc.).
    hostname : Optional[str]
        nom facultatif du hostname (si non renseigné sera complété 
        automatiquement par Artimon).
    environment : Optional[str]
        nom facultatif de l'environnement (si non renseigné sera complété 
        automatiquement par Artimon).
    time_life : int (**deprecated**)
        nombre de jour d'activité du fichier récoltant les métriques. Le
        collecteur Artimon supprime automatiquement les fichiers qui ont 7 
        jours d'inactivité (par défaut 1 jours).
    
    Notes
    -----
    Avec cette classe il n'est pas possible d'utiliser le format "message", 
    il est important de ce référer à "msg" pour le message. De même il n'est 
    pas nécessaire de mettre "msg" dans l'argument format car le contenu du 
    message sera utilisé automatiquement comme argument de la série.

    L'argument time_life n'est plus pris en compte et sera supprimé dans les 
    versions futures.

    Examples
    --------
    Envoi d'un timer mesurant de temps d'exécution d'une fonction et de sans 
    sortie.
    
    >>> import logging
    >>> from time import sleep
    >>> import numpy as np
    >>> from dstk.metrics import ArtimonLocalHandler, timer, sniffer
    >>> 
    >>> # Création du logger
    >>> logging.basicConfig(
    >>>     level = logging.INFO,
    >>>     handlers = [
    >>>         ArtimonLocalHandler(
    >>>             producer = 'NAME.producer',
    >>>             format = ['asctime', 'threadName']
    >>>         )
    >>>     ]
    >>> )
    >>> logger_timer = logging.getLogger('metrics.python.timer')
    >>> logger_sniffer = logging.getLogger('metrics.python.sniffer')
    
    >>> @timer(logger=logger_timer, delta_time=True)
    >>> @sniffer(logger=logger_sniffer)
    >>> def function_test(a=1, b=1, sleep_time=0.5):
    >>>     r"Docstring de la fonction fonction_test"
    >>>     sleep(sleep_time)
    >>>     return a+b
    
    >>> res = [
    >>>     function_test(a=ii, sleep_time=5*np.random.rand())
    >>>     for ii in range(10)
    >>> ]
    >>> res
    [1, 3, 4, 5, 6, 7, 8, 9, 10]
    
    Les données seront envoyés directement vers le fichier 
    date_du_jour.artimon dans le répertoire '/var/run/artimonV2'. C'est le 
    collecteur Artimon qui ce chargera de l'envoie vers Warp10.
    
    See also
    --------
    ArtimonAPIHandler
    """
    def __init__(
        self,
        producer: str,
        path: str='/var/run/artimonV2',
        format: List[str]=['name', 'msg'],
        hostname: Optional[str]=None,
        environment: Optional[str]=None,
        #time_life: int=1
    ) -> None:
        if 'message' in format:
            raise AttributeError(
                "Ne pas utiliser l'attribut 'message' qui est technique, mais "
                "utiliser plutôt l'attribut 'msg'."
            )
        super().__init__(
            format = format,
            producer = producer,
            hostname = hostname,
            environment = environment
        )
        self.path = path
        # self.time_life = time_life
        if not os.path.exists(path):
            raise FileNotFoundError(
                f"Le répertoire {path} n'existe pas."
            )

    def emit(self, record: logging.LogRecord) -> None:
        self._asctime(record)
        self._engine(record, 'artimon')
        file_name = f"{dt.now().strftime('%Y%m%d')}"
        with open(f"{file_name}.tmp", 'w') as file:
            file.write(self.parser(record)+'\n')
        os.rename(f"{file_name}.tmp", f"{file_name}.artimon")
        # A remove si plus necessaire
        # file_name = sorted(
        #     filter(
        #         lambda x: x.endswith('.artimon'),
        #         os.listdir()
        #     ),
        #     reverse = True
        # )
        # if len(file_name) > 0:
        #     file_name = file_name[0]
        #     delta_j = (
        #         dt.now().date()
        #         -dt.strptime(file_name.split('.')[0], '%Y%m%d').date()
        #     ).days
        #     if delta_j > self.time_life:
        #         file_name = f"{dt.now().strftime('%Y%m%d')}.artimon"
        # else:
        #     file_name = f"{dt.now().strftime('%Y%m%d')}.artimon"
        # with open(file_name, 'a') as file:
        #     file.write(self.parser(record)+'\n')

class Warp10Handler(HandlerBase):
    r"""
    Classe de base permettant de créer un handler à partir d'une liste de 
    Handler vers une base interfacée à Warp10.
    """
    def __init__(
        self,
        token: str,
        host: str='127.0.0.1',
        port: str='8080',
        field_value: str='value',
        format: List[str]=['name', 'msg'],
        api_version: str='v0'
    ) -> None:
        try:
            global requests
            import requests
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Le module requests n'est pas installé : l'instruction "
                "import requests renvoit une erreur "
                "ModuleNotFoundError. Rajouter requests dans les "
                "packages à installer dans le fichier environment.yml du "
                "projet."
            )
        if 'message' in format:
            raise AttributeError(
                "Ne pas utiliser l'attribut 'message' qui est technique, mais "
                "utiliser plutôt l'attribut 'msg'."
            )
        if not isinstance(field_value, dict):
            raise AttributeError(
                "field_value doit être un dictionnaire avec comme clef le "
                "nom du logger et comme valeur le nom du champ ou la liste "
                "si plusieur champs."
            )
        super().__init__(format = format)
        self.headers = {
         'X-Warp10-Token': token
        }
        self.host = host
        self.port = port
        self.field_value = field_value
        self.api_version = api_version
        self.url = (
            f"http://{self.host}:{self.port}/api/{self.api_version}/update"
        )

    def emit(self, record: logging.LogRecord) -> None:
        self._asctime(record)
        self._engine(record, 'warp10')
        gts = GTS(record, self.field_value, self.format)
        response = requests.post(
            self.url,
            headers = self.headers,
            data = gts.commit()
        )
        if response.status_code != 200:
            raise IOError(
                "Le retour du puch vers Warp10 à renvoyer une erreur. L'API "
                f"a renvoyer le code {response.status_code}."
            )

class GTS:
    r"""
    Classe modélisant un objet GTS (Geo Time Series) Warp10.
    """
    def __init__(
        self,
        record: logging.LogRecord,
        dict_value: list,
        format: List[str]
    ) -> None:
        self.msg = record.msg.copy()
        field_value = dict_value[record.name]
        if isinstance(field_value, str):
            field_value = [field_value]
        if not all([ii in self.msg for ii in field_value]):
            raise IOError(
                "Le ou l'un des champs de field_value n'est ou ne sont pas "
                "présent(s) dans les données du message envoyées par le "
                f"logger qui sont :\n{', '.join(self.msg.keys())}."
            )
        self.format = format
        self.ts = f"{str(record.created).replace('.', '')[:16]:0<16}"
        self.name = record.name
        self.value = [self.msg.pop(ii) for ii in field_value]
        if 'created' in self.format:
            self.format.remove('created')
        if 'name' in self.format:
            self.format.remove('name')
        if 'msg' in self.format:
            self.format.remove('msg')
        log_att = record.__dict__
        self.log_att = list(map(lambda x: log_att[x], self.format))

    def commit(self) -> str:
        query = (
            f"{self.ts}// {self.name}"
            f"{{{','.join([f'{kk}={vv}' for kk, vv in zip(self.format, self.log_att)])}}}"
            f"{{{','.join([f'{kk}={vv}' for kk, vv in self.msg.items()])}}} "
        )
        if len(self.value) == 1:
            value = (
                f"'{self.value[0]}'"
                if isinstance(self.value[0], str) else
                f"{self.value[0]}"
            )
        else:
            value = (
                f"""[ {' '.join([(f"'{ii}'" if isinstance(ii, str) else f'{ii}') 
                for ii in self.value])} ]"""
            )
        return query+value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Décorateurs deprecated
^^^^^^^^^^^^^^^^^^^^^^

Created on Mon Nov 23 11:29:42 2020

@author: Cyrile Delestre
"""

from functools import singledispatch, wraps
from warnings import warn

@singledispatch
def deprecated(fun, *args, **kwargs):
    """
    Décorateur indiquant que la fonction est dépréciée.
    
    Examples
    --------
    >>> from dstk.utils.deprecated import deprecated
    
    >>> @deprecated
    >>> def fun():
    >>>     print("Hello world!")
    
    >>> fun()
    :1: DeprecationWarning: La fonction fun est dépreciée !
    """
    @wraps(fun)
    def wrapper(*args, **kwargs):
        warn(
            f"La fonction {fun.__name__} est dépreciée !",
            DeprecationWarning
        )
        return fun(*args, **kwargs)
    return wrapper

@deprecated.register(str)
def _deprecated(msg: str, *args, **kwargs):
    """
    Décorateur indiquant que la fonction est dépréciée et affiche un message 
    entré par l'utilisateur.
    
    Parameters
    ----------
    msg : str
        message à transmettre dans le warning
    
    Examples
    --------
    >>> from dstk.utils.deprecated import deprecated
    
    >>> @deprecated("Fonction plus valide dans la prochaine version.")
    >>> def fun():
    >>>     print("Hello world!")
    
    >>> fun()
    :1: DeprecationWarning: La fonction fun est dépreciée !
    Fonction plus valide dans la prochaine version.
    """
    def decorator(fun):
        @wraps(fun)
        def wrapper(*args, **kwargs):
            warn(
                f"La fonction {fun.__name__} est dépreciée !\n{msg}",
                DeprecationWarning
            )
            return fun(*args, **kwargs)
        return wrapper
    return decorator



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Fonction set_params_deep
^^^^^^^^^^^^^^^^^^^^^^^^

Created on Mon Nov 23 13:23:47 2020

@author: Cyrile Delestre
"""

from warnings import warn

from sklearn.base import BaseEstimator

def set_params_deep(model: BaseEstimator, warnning: bool=True, **params):
    r"""
    Permet de changer l'ensemble de paramètres à partir d'un certain niveau 
    de l'attribution des noms de variables dans le standard Scikit-Learn 
    "nv_1__niv_2__niv_3__param".
    
    Parameters
    ----------
    model : BaseEstimator
        model Scikit-Learn (Pipeline, algo ML, grid search, etc.).
    warnning : bool
        affiche un warning si le paramètre recherché n'est pas présent dans 
        le modèle
    **params :
        paramètres à modifier au format key__params. Si on souhaite changer 
        tous les paramètres qui possède le même nom alors juste mettre params.
    
    Examples
    --------
    >>> from sklearn.pipeline import Pipeline
    >>> ...
    >>> from dstk.utils import set_params_deep
    
    >>> ...
    >>> 
    >>> pipe = Pipeline(...)
    >>> 
    >>> # Par exemple si on souhaite changer tous les n_jobs, max_depth et
    >>> # verbose dans l'ensemble de la Pipeline pipe
    >>> set_params_deep(
    >>>     pipe,
    >>>     n_jobs=1,
    >>>     verbose=False,
    >>>     niv_1__niv_2__max_depth=42
    >>> )
    """
    if not (hasattr(model, 'get_params') and hasattr(model, 'set_params')):
        raise AttributeError(
            "La modèle doit être un objet de type Scikit-Learn et doit être "
            "doté d'une méthode get_params et set_params. L'objet model est "
            f"ici de type {type(model)}."
        )
    list_params = model.get_params().keys()
    new_params = dict()
    for kk in params:
        keys = kk.split('__')
        if len(keys) == 1:
            filt = list(
                filter(
                    lambda x: x.split('__')[-1] == kk,
                    list_params
                )
            )
        else:
            filt = list(
                filter(
                    lambda x: (
                        len(x.split('__')) >= len(keys)
                        and
                        all(
                            [
                                ii == tt
                                for ii, tt in zip(x.split('__'), keys[:-1])
                            ]
                        )
                        and x.split('__')[-1] == keys[-1]
                    ),
                    list_params
                )
            )
        if warnning and len(filt) == 0:
            if len(keys) == 1:
                warn(
                    f"Le paramètre {kk} n'est pas présent dans le modèle."
                )
            else:
                warn(
                    f"La clef {'__'.join(keys[:-1])} avec le paramètre "
                    f"{keys[-1]} n'est pas présent dans le modèle."
                )
        if len(filt) > 0:
            new_params.update({param: params[kk] for param in filt})
    if len(new_params) > 0:
        model.set_params(**new_params)
